__all__ = ['entropy_vn', 'entropy_linear', 'entropy_mutual', 'negativity',
           'concurrence', 'entropy_conditional', 'entangling_power',
           'entropy_relative']

from numpy import conj, e, inf, imag, inner, real, sort, sqrt
from numpy.lib.scimath import log, log2
from . import (ptrace, ket2dm, tensor, sigmay, partial_transpose,
               expand_operator)
from .core import data as _data


def entropy_vn(rho, base=e, sparse=False):
    """
    Von-Neumann entropy of density matrix

    Parameters
    ----------
    rho : qobj
        Density matrix.
    base : {e, 2}, default: e
        Base of logarithm.
    sparse : bool, default: False
        Use sparse eigensolver.

    Returns
    -------
    entropy : float
        Von-Neumann entropy of `rho`.

    Examples
    --------
    >>> rho=0.5*fock_dm(2,0)+0.5*fock_dm(2,1)
    >>> entropy_vn(rho,2)
    1.0

    """
    if rho.type == 'ket' or rho.type == 'bra':
        rho = ket2dm(rho)
    vals = rho.eigenenergies(sparse=sparse)
    nzvals = vals[vals != 0]
    if base == 2:
        logvals = log2(nzvals)
    elif base == e:
        logvals = log(nzvals)
    else:
        raise ValueError("Base must be 2 or e.")
    return float(real(-sum(nzvals * logvals)))


def entropy_linear(rho):
    """
    Linear entropy of a density matrix.

    Parameters
    ----------
    rho : qobj
        sensity matrix or ket/bra vector.

    Returns
    -------
    entropy : float
        Linear entropy of rho.

    Examples
    --------
    >>> rho=0.5*fock_dm(2,0)+0.5*fock_dm(2,1)
    >>> entropy_linear(rho)
    0.5

    """
    if rho.type == 'ket' or rho.type == 'bra':
        rho = ket2dm(rho)
    return float(real(1.0 - (rho ** 2).tr()))


def concurrence(rho):
    """
    Calculate the concurrence entanglement measure for a two-qubit state.

    Parameters
    ----------
    state : qobj
        Ket, bra, or density matrix for a two-qubit state.

    Returns
    -------
    concur : float
        Concurrence

    References
    ----------

    .. [1] `https://en.wikipedia.org/wiki/Concurrence_(quantum_computing)`

    """
    if rho.isket and rho.dims != [[2, 2], [1, 1]]:
        raise Exception("Ket must be tensor product of two qubits.")

    elif rho.isbra and rho.dims != [[1, 1], [2, 2]]:
        raise Exception("Bra must be tensor product of two qubits.")

    elif rho.isoper and rho.dims != [[2, 2], [2, 2]]:
        raise Exception("Density matrix must be tensor product of two qubits.")

    if rho.isket or rho.isbra:
        rho = ket2dm(rho)

    sysy = tensor(sigmay(), sigmay())

    rho_tilde = (rho * sysy) * (rho.conj() * sysy)

    evals = rho_tilde.eigenenergies()

    # abs to avoid problems with sqrt for very small negative numbers
    evals = abs(sort(real(evals)))

    lsum = sqrt(evals[3]) - sqrt(evals[2]) - sqrt(evals[1]) - sqrt(evals[0])

    return max(0, lsum)


def negativity(rho, subsys, method='tracenorm', logarithmic=False):
    """
    Compute the negativity for a multipartite quantum system described
    by the density matrix rho. The subsys argument is an index that
    indicates which system to compute the negativity for.

    .. note::

        Experimental.
    """
    mask = [idx == subsys for idx, n in enumerate(rho.dims[0])]
    rho_pt = partial_transpose(rho, mask)

    if method == 'tracenorm':
        N = ((rho_pt.dag() * rho_pt).sqrtm().tr().real - 1)/2.0
    elif method == 'eigenvalues':
        l = rho_pt.eigenenergies()
        N = ((abs(l)-l)/2).sum()
    else:
        raise ValueError("Unknown method %s" % method)

    if logarithmic:
        return log2(2 * N + 1)
    else:
        return N


def entropy_mutual(rho, selA, selB, base=e, sparse=False):
    """
    Calculates the mutual information S(A:B) between selection
    components of a system density matrix.

    Parameters
    ----------
    rho : qobj
        Density matrix for composite quantum systems
    selA : int/list
        `int` or `list` of first selected density matrix components.
    selB : int/list
        `int` or `list` of second selected density matrix components.
    base : {e, 2}, default: e
        Base of logarithm.
    sparse : bool, default: False
        Use sparse eigensolver.

    Returns
    -------
    ent_mut : float
       Mutual information between selected components.

    """
    if isinstance(selA, int):
        selA = [selA]
    if isinstance(selB, int):
        selB = [selB]
    if rho.type != 'oper':
        raise TypeError("Input must be a density matrix.")
    if (len(selA) + len(selB)) != len(rho.dims[0]):
        raise TypeError("Number of selected components must match " +
                        "total number.")

    rhoA = ptrace(rho, selA)
    rhoB = ptrace(rho, selB)
    out = (entropy_vn(rhoA, base, sparse=sparse) +
           entropy_vn(rhoB, base, sparse=sparse) -
           entropy_vn(rho, base, sparse=sparse))
    return out


def entropy_relative(rho, sigma, base=e, sparse=False, tol=1e-12):
    """
    Calculates the relative entropy S(rho||sigma) between two density
    matrices.

    Parameters
    ----------
    rho : :class:`.Qobj`
        First density matrix (or ket which will be converted to a density
        matrix).
    sigma : :class:`.Qobj`
        Second density matrix (or ket which will be converted to a density
        matrix).
    base : {e, 2}, default: e
        Base of logarithm. Defaults to e.
    sparse : bool, default: False
        Flag to use sparse solver when determining the eigenvectors
        of the density matrices. Defaults to False.
    tol : float, default: 1e-12
        Tolerance to use to detect 0 eigenvalues or dot producted between
        eigenvectors. Defaults to 1e-12.

    Returns
    -------
    rel_ent : float
        Value of relative entropy. Guaranteed to be greater than zero
        and should equal zero only when rho and sigma are identical.

    Examples
    --------

    First we define two density matrices:

    >>> rho = qutip.ket2dm(qutip.ket("00"))
    >>> sigma = rho + qutip.ket2dm(qutip.ket("01"))
    >>> sigma = sigma.unit()

    Then we calculate their relative entropy using base 2 (i.e. ``log2``)
    and base e (i.e. ``log``).

    >>> qutip.entropy_relative(rho, sigma, base=2)
    1.0
    >>> qutip.entropy_relative(rho, sigma)
    0.6931471805599453

    References
    ----------

    See Nielsen & Chuang, "Quantum Computation and Quantum Information",
    Section 11.3.1, pg. 511 for a detailed explanation of quantum relative
    entropy.
    """
    if rho.isket:
        rho = ket2dm(rho)
    if sigma.isket:
        sigma = ket2dm(sigma)
    if not rho.isoper or not sigma.isoper:
        raise TypeError("Inputs must be density matrices.")
    if rho.dims != sigma.dims:
        raise ValueError("Inputs must have the same shape and dims.")
    if base == 2:
        log_base = log2
    elif base == e:
        log_base = log
    else:
        raise ValueError("Base must be 2 or e.")
    # S(rho || sigma) = sum_i(p_i log p_i) - sum_ij(p_i P_ij log q_i)
    #
    # S is +inf if the kernel of sigma (i.e. svecs[svals == 0]) has non-trivial
    # intersection with the support of rho (i.e. rvecs[rvals != 0]).
    rvals, rvecs = _data.eigs(rho.data, rho.isherm, True)
    rvecs = rvecs.to_array().T
    if any(abs(imag(rvals)) >= tol):
        raise ValueError("Input rho has non-real eigenvalues.")
    rvals = real(rvals)
    svals, svecs = _data.eigs(sigma.data, sigma.isherm, True)
    svecs = svecs.to_array().T
    if any(abs(imag(svals)) >= tol):
        raise ValueError("Input sigma has non-real eigenvalues.")
    svals = real(svals)
    # Calculate inner products of eigenvectors and return +inf if kernel
    # of sigma overlaps with support of rho.
    P = abs(inner(rvecs, conj(svecs))) ** 2
    if (rvals >= tol) @ (P >= tol) @ (svals < tol):
        return inf
    # Avoid -inf from log(0) -- these terms will be multiplied by zero later
    # anyway
    svals[abs(svals) < tol] = 1
    nzrvals = rvals[abs(rvals) >= tol]
    # Calculate S
    S = nzrvals @ log_base(nzrvals) - rvals @ P @ log_base(svals)
    # the relative entropy is guaranteed to be >= 0, so we clamp the
    # calculated value to 0 to avoid small violations of the lower bound.
    return max(0, S)


def entropy_conditional(rho, selB, base=e, sparse=False):
    """
    Calculates the conditional entropy :math:`S(A|B)=S(A,B)-S(B)`
    of a selected density matrix component.

    Parameters
    ----------
    rho : qobj
        Density matrix of composite object
    selB : int/list
        Selected components for density matrix B
    base : {e, 2}, default: e
        Base of logarithm.
    sparse : bool, default: False
        Use sparse eigensolver.

    Returns
    -------
    ent_cond : float
        Value of conditional entropy

    """
    if rho.type != 'oper':
        raise TypeError("Input must be density matrix.")
    if isinstance(selB, int):
        selB = [selB]
    B = ptrace(rho, selB)
    out = (entropy_vn(rho, base, sparse=sparse) -
           entropy_vn(B, base, sparse=sparse))
    return out


def participation_ratio(rho):
    """
    Returns the effective number of states for a density matrix.

    The participation is unity for pure states, and maximally N,
    where N is the Hilbert space dimensionality, for completely
    mixed states.

    Parameters
    ----------
    rho : qobj
        Density matrix

    Returns
    -------
    pr : float
        Effective number of states in the density matrix

    """
    if rho.type == 'ket' or rho.type == 'bra':
        return 1.0
    else:
        return 1.0 / (rho ** 2).tr()


def entangling_power(U):
    """
    Calculate the entangling power of a two-qubit gate U, which
    is zero of nonentangling gates and 2/9 for maximally entangling gates.

    Parameters
    ----------
    U : qobj
        Qobj instance representing a two-qubit gate.

    Returns
    -------
    ep : float
        The entanglement power of U (real number between 0 and 2/9)

    References:

        Explorations in Quantum Computing, Colin P. Williams (Springer, 2011)
    """

    if not U.isoper:
        raise Exception("U must be an operator.")

    if U.dims != [[2, 2], [2, 2]]:
        raise Exception("U must be a two-qubit gate.")

    from qutip.core.gates import swap
    swap13 =  expand_operator(swap(), [2, 2, 2, 2], [1, 3])
    a = tensor(U, U).dag() * swap13 * tensor(U, U) * swap13
    Uswap = swap() * U
    b = tensor(Uswap, Uswap).dag() * swap13 * tensor(Uswap, Uswap) * swap13

    return 5.0/9 - 1.0/36 * (a.tr() + b.tr()).real



#cython: language_level=3
#cython: c_api_binop_methods=True

import numpy as np
cimport numpy as cnp
import qutip.core.data as _data
from qutip.settings import settings

__all__ = [
    'idxint_size', 'idxint_dtype', 'Data', 'EfficiencyWarning',
]

if _idxint_size == 32:
    idxint_dtype = np.int32
    idxint_DTYPE = cnp.NPY_INT32
else:
    idxint_dtype = np.int64
    idxint_DTYPE = cnp.NPY_INT64

idxint_size = _idxint_size

# As this is an abstract base class with C entry points, we have to explicitly
# stub out methods since we can't mark them as abstract.
cdef class Data:
    def __init__(self, shape):
        self.shape = shape

    cpdef object to_array(self):
        raise NotImplementedError

    cpdef double complex trace(self):
        return NotImplementedError

    cpdef Data adjoint(self):
        raise NotImplementedError

    cpdef Data conj(self):
        raise NotImplementedError

    cpdef Data transpose(self):
        raise NotImplementedError

    cpdef Data copy(self):
        raise NotImplementedError

    def __add__(left, right):
        if isinstance(left, Data) and isinstance(right, Data):
            return _data.add(left, right)
        return NotImplemented

    def __sub__(left, right):
        if isinstance(left, Data) and isinstance(right, Data):
            return _data.sub(left, right)
        return NotImplemented

    def __matmul__(left, right):
        if isinstance(left, Data) and isinstance(right, Data):
            return _data.matmul(left, right)
        return NotImplemented

    def __mul__(left, right):
        data, number = (left, right) if isinstance(left, Data) else (right, left)
        try:
            return _data.mul(data, number)
        except TypeError:
            return NotImplemented

    def __truediv__(left, right):
        data, number = (left, right) if isinstance(left, Data) else (right, left)
        try:
            return _data.mul(data, 1/number)
        except TypeError:
            return NotImplemented

    def __neg__(self):
        return _data.neg(self)

    def __eq__(left, right):
        if not (isinstance(left, Data) and isinstance(right, Data)):
            return NotImplemented
        if (
            left.shape[0] == right.shape[0]
            and left.shape[1] == right.shape[1]
        ):
            return _data.iszero(_data.sub(left, right), settings.core['atol'])
        return False


class EfficiencyWarning(Warning):
    pass



#cython: language_level=3
from qutip.core.cy.coefficient cimport Coefficient


cdef class RateShiftCoefficient(Coefficient):
    cdef:
        Coefficient [:] coeffs

    cpdef double as_double(self, double t) except *


cdef class SqrtRealCoefficient(Coefficient):
    cdef:
        Coefficient base



#cython: language_level=3
#cython: boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True

import numbers

import numpy as np

cimport numpy as cnp
cimport cython

from qutip.core.data cimport csr, dense, idxint, CSR, Dense, Data, Dia, dia
from qutip.core.data.base import idxint_dtype
from qutip.settings import settings

cnp.import_array()

__all__ = [
    'ptrace', 'ptrace_csr', 'ptrace_dense', 'ptrace_csr_dense', 'ptrace_dia',
]

cdef tuple _parse_inputs(object dims, object sel, tuple shape):
    cdef Py_ssize_t i

    dims = np.atleast_1d(dims).astype(idxint_dtype).ravel()
    sel = np.atleast_1d(sel).astype(idxint_dtype)
    sel.sort()

    if shape[0] != shape[1]:
        raise ValueError("ptrace is only defined for square density matrices")

    if shape[0] != np.prod(dims, dtype=int):
        raise ValueError(f"the input matrix shape, {shape} and the"
                         f" dimension argument, {dims}, are not compatible.")
    if sel.ndim != 1:
        raise ValueError("Selection must be one-dimensional")

    if any(d < 1 for d in dims):
        raise ValueError("dimensions must be greated than zero but where"
                         f" dims={dims}.")

    for i in range(sel.shape[0]):
        if sel[i] < 0 or sel[i] >= dims.size:
            raise IndexError("Invalid selection index in ptrace.")
        if i > 0 and sel[i] == sel[i - 1]:
            raise ValueError("Duplicate selection index in ptrace.")

    return dims, sel

cdef idxint _populate_tensor_table(dims, sel, idxint[:, ::1] tensor_table) except -1:
    """
    Populate the helper structure `tensor_table`.  Returns the size (number of
    rows and number of columns) of the matrix which will be output.
    """
    cdef size_t ii
    cdef idxint[::1] _dims = np.asarray(dims, dtype=idxint_dtype).ravel()
    cdef size_t num_dims = _dims.shape[0]
    cdef idxint factor_tensor=1, factor_keep=1, factor_trace=1
    cdef idxint[::1] _sel = np.asarray(sel, dtype=idxint_dtype)
    for ii in range(_sel.shape[0]):
        if _sel[ii] < 0 or _sel[ii] >= num_dims:
            raise TypeError("Invalid selection index in ptrace.")
    for ii in range(num_dims - 1,-1,-1):
        tensor_table[ii, 0] = factor_tensor
        factor_tensor *= _dims[ii]
        if _in(ii, _sel):
            tensor_table[ii, 1] = factor_keep
            factor_keep *= _dims[ii]
        else:
            tensor_table[ii, 2]  = factor_trace
            factor_trace *= _dims[ii]
    return factor_keep


cdef bint _in(idxint val, idxint[::1] vec):
    cdef int ii
    for ii in range(vec.shape[0]):
        if val == vec[ii]:
            return True
    return False


cdef inline void _i2_k_t(idxint N, idxint[:, ::1] tensor_table, idxint out[2]):
    # indices determining function for ptrace
    cdef size_t ii
    cdef idxint t1, t2
    out[0] = out[1] = 0
    for ii in range(tensor_table.shape[0]):
        t1 = tensor_table[ii, 0]
        t2 = N / t1
        N = N % t1
        out[0] += tensor_table[ii, 1] * t2
        out[1] += tensor_table[ii, 2] * t2


cpdef CSR ptrace_csr(CSR matrix, object dims, object sel):
    dims, sel = _parse_inputs(dims, sel, matrix.shape)

    if len(sel) == len(dims):
        return matrix.copy()
    cdef idxint[:, ::1] tensor_table = np.zeros((dims.shape[0], 3), dtype=idxint_dtype)
    cdef idxint size
    size = _populate_tensor_table(dims, sel, tensor_table)
    cdef size_t p=0, nnz=csr.nnz(matrix), row, ptr
    cdef idxint pos_c[2]
    cdef idxint pos_r[2]
    cdef cnp.ndarray[double complex, ndim=1, mode='c'] new_data = np.zeros(nnz, dtype=complex)
    cdef cnp.ndarray[idxint, ndim=1, mode='c'] new_col = np.zeros(nnz, dtype=idxint_dtype)
    cdef cnp.ndarray[idxint, ndim=1, mode='c'] new_row = np.zeros(nnz, dtype=idxint_dtype)
    cdef double tol = 0
    if settings.core['auto_tidyup']:
        tol = settings.core['auto_tidyup_atol']
    for row in range(matrix.shape[0]):
        for ptr in range(matrix.row_index[row], matrix.row_index[row + 1]):
            _i2_k_t(matrix.col_index[ptr], tensor_table, pos_c)
            _i2_k_t(row, tensor_table, pos_r)
            if pos_c[1] == pos_r[1]:
                new_data[p] = matrix.data[ptr]
                new_row[p] = pos_r[0]
                new_col[p] = pos_c[0]
                p += 1
    return csr.from_coo_pointers(&new_row[0], &new_col[0], &new_data[0],
                                 size, size, p, tol)


def ptrace_dia(matrix, dims, sel):
    if len(sel) == len(dims):
        return matrix.copy()
    dims, sel = _parse_inputs(dims, sel, matrix.shape)
    mat = matrix.as_scipy()
    cdef idxint[:, ::1] tensor_table = np.zeros((dims.shape[0], 3), dtype=idxint_dtype)
    cdef idxint pos_row[2]
    cdef idxint pos_col[2]
    size = _populate_tensor_table(dims, sel, tensor_table)
    data = {}
    for i, offset in enumerate(mat.offsets):
        start = max(0, offset)
        end = min(matrix.shape[0] + offset, matrix.shape[1])
        for col in range(start, end):
            _i2_k_t(col - offset, tensor_table, pos_row)
            _i2_k_t(col, tensor_table, pos_col)
            if pos_row[1] == pos_col[1]:
                new_offset = pos_col[0] - pos_row[0]
                if new_offset not in data:
                    data[new_offset] = np.zeros(size, dtype=complex)
                data[new_offset][pos_col[0]] += mat.data[i, col]

    if len(data) == 0:
        return dia.zeros(size, size)
    offsets = np.array(list(data.keys()), dtype=idxint_dtype)
    data = np.array(list(data.values()), dtype=complex)
    out = Dia((data, offsets), shape=(size, size), copy=False)
    out = dia.clean_dia(out, True)
    return out


cpdef Dense ptrace_csr_dense(CSR matrix, object dims, object sel):
    dims, sel = _parse_inputs(dims, sel, matrix.shape)

    if len(sel) == len(dims):
        return dense.from_csr(matrix)
    cdef idxint[:, ::1] tensor_table = np.zeros((dims.shape[0], 3), dtype=idxint_dtype)
    cdef idxint size
    size = _populate_tensor_table(dims, sel, tensor_table)
    cdef size_t ii, jj
    cdef idxint pos_c[2]
    cdef idxint pos_r[2]
    cdef Dense out = dense.zeros(size, size, fortran=False)
    for ii in range(matrix.shape[0]):
        for jj in range(matrix.row_index[ii], matrix.row_index[ii+1]):
            _i2_k_t(matrix.col_index[jj], tensor_table, pos_c)
            _i2_k_t(ii, tensor_table, pos_r)
            if pos_c[1] == pos_r[1]:
                out.data[pos_r[0]*size + pos_c[0]] += matrix.data[jj]
    return out


cpdef Dense ptrace_dense(Dense matrix, object dims, object sel):
    dims, sel = _parse_inputs(dims, sel, matrix.shape)

    if len(sel) == len(dims):
        return matrix.copy()
    nd = dims.shape[0]
    dkeep = [dims[x] for x in sel]
    qtrace = list(set(np.arange(nd)) - set(sel))
    dtrace = [dims[x] for x in qtrace]
    dims = list(dims)
    sel = list(sel)
    rhomat = np.trace(matrix.as_ndarray()
                      .reshape(dims + dims)
                      .transpose(qtrace + [nd + q for q in qtrace] +
                                 sel + [nd + q for q in sel])
                      .reshape([np.prod(dtrace, dtype=int),
                                np.prod(dtrace, dtype=int),
                                np.prod(dkeep, dtype=int),
                                np.prod(dkeep, dtype=int)]))
    return dense.fast_from_numpy(rhomat)


from .dispatch import Dispatcher as _Dispatcher
import inspect as _inspect

ptrace = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('matrix', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('dims', _inspect.Parameter.POSITIONAL_OR_KEYWORD),
        _inspect.Parameter('sel', _inspect.Parameter.POSITIONAL_OR_KEYWORD),
    ]),
    name='ptrace',
    module=__name__,
    inputs=('matrix',),
    out=True,
)
ptrace.__doc__ =\
    """
    Compute the partial trace of this matrix, leaving the subspaces whose
    indices are in `sel`.  This is only defined for square density matrices,
    and always returns a density matrix.

    The order of the indices in `sel` do not matter; the output matrix will
    always have the selected subspaces in the same order that they are in the
    input matrix.

    For example, if the input is the matrix backing a `Qobj` with
    `dims = [[2, 3, 4], [2, 3, 4]]`, then the output of
        ptrace(data, [2, 3, 4], [0, 2])
    will be a matrix with effective dimensions `[[2, 4], [2, 4]]`.

    Parameters
    ----------
    matrix : Data
        The density matrix to be partially traced.

    dims : array_like of integer
        The dimensions of the subspaces.  This is most likely a 1D list, and
        since this is only defined on square matrices, there is no need to pass
        before the left and right sides.  Typically the input matrix will be
        taken from a `Qobj`, and then this parameter will be `Qobj.dims[0]`.

    sel : integer or array_like of integer
        The indices of the subspaces which should be _kept_.
    """
ptrace.add_specialisations([
    (CSR, CSR, ptrace_csr),
    (CSR, Dense, ptrace_csr_dense),
    (Dense, Dense, ptrace_dense),
    (Dia, Dia, ptrace_dia),
], _defer=True)

del _inspect, _Dispatcher



"""
Functions for visualizing results of quantum dynamics simulations,
visualizations of quantum states and processes.
"""

__all__ = ['plot_wigner_sphere', 'hinton', 'sphereplot',
           'matrix_histogram', 'plot_energy_levels', 'plot_fock_distribution',
           'plot_wigner', 'plot_expectation_values',
           'plot_spin_distribution', 'complex_array_to_rgb',
           'plot_qubism', 'plot_schmidt']

import warnings
import itertools as it
import numpy as np
from numpy import pi, array, sin, cos, angle, log2, sqrt

from packaging.version import parse as parse_version

from . import (
    Qobj, isket, ket2dm, tensor, vector_to_operator, to_super, settings
)
from .core.dimensions import flatten
from .core.superop_reps import _to_superpauli, isqubitdims
from .wigner import wigner
from .matplotlib_utilities import complex_phase_cmap

try:
    import matplotlib.pyplot as plt
    import matplotlib as mpl
    import matplotlib.animation as animation
    from matplotlib import cm
    from mpl_toolkits.mplot3d import Axes3D

    # Define a custom _axes3D function based on the matplotlib version.
    # The auto_add_to_figure keyword is new for matplotlib>=3.4.
    if parse_version(mpl.__version__) >= parse_version('3.4'):
        def _axes3D(fig, *args, **kwargs):
            ax = Axes3D(fig, *args, auto_add_to_figure=False, **kwargs)
            return fig.add_axes(ax)
    else:
        def _axes3D(*args, **kwargs):
            return Axes3D(*args, **kwargs)
except:
    pass


def _cyclic_cmap():
    if settings.colorblind_safe:
        return cm.twilight
    else:
        return complex_phase_cmap()


def _diverging_cmap():
    if settings.colorblind_safe:
        return cm.seismic
    else:
        return cm.RdBu


def _sequential_cmap():
    if settings.colorblind_safe:
        return cm.cividis
    else:
        return cm.jet


def _is_fig_and_ax(fig, ax, projection='2d'):
    if fig is None:
        if ax is None:
            fig = plt.figure()
            if projection == '2d':
                ax = fig.add_subplot(1, 1, 1)
            else:
                ax = _axes3D(fig)
        else:
            fig = ax.get_figure()
    else:
        if ax is None:
            if projection == '2d':
                ax = fig.add_subplot(1, 1, 1)
            else:
                ax = _axes3D(fig)

    return fig, ax


def _set_ticklabels(ax, ticklabels, ticks, axis, fontsize=14):
    if len(ticks) != len(ticklabels):
        raise ValueError(
            f"got {len(ticklabels)} ticklabels but needed {len(ticks)}"
        )
    if axis == 'x':
        ax.set_xticks(ticks)
        ax.set_xticklabels(ticklabels, fontsize=fontsize)
    elif axis == 'y':
        ax.set_yticks(ticks)
        ax.set_yticklabels(ticklabels, fontsize=fontsize)
    else:
        raise ValueError(
            "axis must be either 'x' or 'y'"
        )


def _equal_shape(matrices):
    first_shape = matrices[0].shape

    text = "All inputs should have the same shape."
    if not all(matrix.shape == first_shape for matrix in matrices):
        raise ValueError(text)


def plot_wigner_sphere(wigner, reflections=False, *, cmap=None,
                       colorbar=True, fig=None, ax=None):
    """Plots a coloured Bloch sphere.

    Parameters
    ----------
    wigner : a wigner transformation
        The wigner transformation at `steps` different theta and phi.

    reflections : bool, default: False
        If the reflections of the sphere should be plotted as well.

    cmap : a matplotlib colormap instance, optional
        Color map to use when plotting.

    colorbar : bool, default: True
        Whether (True) or not (False) a colorbar should be attached.

    fig : a matplotlib Figure instance, optional
        The Figure canvas in which the plot will be drawn.

    ax : a matplotlib axes instance, optional
        The ax context in which the plot will be drawn.

    Returns
    -------
    fig, output : tuple
        A tuple of the matplotlib figure and the axes instance or animation
        instance used to produce the figure.

    Notes
    -----
    Special thanks to Russell P Rundle for writing this function.
    """

    fig, ax = _is_fig_and_ax(fig, ax, projection='3d')

    if not isinstance(wigner, list):
        wigners = [wigner]
    else:
        wigners = wigner

    _equal_shape(wigners)

    wigner_max = np.real(np.amax(np.abs(wigners[0])))
    for wigner in wigners:
        wigner_max = max(np.real(np.amax(np.abs(wigner))), wigner_max)

    norm = mpl.colors.Normalize(-wigner_max, wigner_max)

    if cmap is None:
        cmap = _diverging_cmap()

    artist_list = list()
    for wigner in wigners:
        steps = len(wigner)
        theta = np.linspace(0, np.pi, steps)
        phi = np.linspace(0, 2 * np.pi, steps)
        x = np.outer(np.sin(theta), np.cos(phi))
        y = np.outer(np.sin(theta), np.sin(phi))
        z = np.outer(np.cos(theta), np.ones(steps))
        wigner = np.real(wigner)

        artist = list()
        # Plot coloured Bloch sphere:
        artist.append(ax.plot_surface(x, y, z, facecolors=cmap(norm(wigner)),
                                      rcount=steps, ccount=steps, linewidth=0,
                                      zorder=0.5, antialiased=None))

        if reflections:
            side_color = cmap(norm(wigner[0:steps, 0:steps]))

            # Plot bottom reflection:
            artist.append(ax.plot_surface(x[0:steps, 0:steps],
                                          y[0:steps, 0:steps],
                                          -1.5*np.ones((steps, steps)),
                                          facecolors=side_color,
                                          rcount=steps/2, ccount=steps/2,
                                          linewidth=0, zorder=0.5,
                                          antialiased=False))

            # Plot side reflection:
            artist.append(ax.plot_surface(-1.5*np.ones((steps, steps)),
                                          y[0:steps, 0:steps],
                                          z[0:steps, 0:steps],
                                          facecolors=side_color,
                                          rcount=steps/2, ccount=steps/2,
                                          linewidth=0, zorder=0.5,
                                          antialiased=False))

            # Plot back reflection:
            artist.append(ax.plot_surface(x[0:steps, 0:steps],
                                          1.5*np.ones((steps, steps)),
                                          z[0:steps, 0:steps],
                                          facecolors=side_color,
                                          rcount=steps/2, ccount=steps/2,
                                          linewidth=0, zorder=0.5,
                                          antialiased=False))
        artist_list.append(artist)

    if len(wigners) == 1:
        output = ax
    else:
        output = animation.ArtistAnimation(fig, artist_list, interval=50,
                                           blit=True, repeat_delay=1000)

    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")

    # Create colourbar:
    if colorbar:
        cax, kw = mpl.colorbar.make_axes(ax, shrink=0.75, pad=.1)
        mpl.colorbar.ColorbarBase(cax, norm=norm, cmap=cmap)

    return fig, output


# Adopted from the SciPy Cookbook.
def _blob(x, y, w, w_max, area, color_fn, ax=None):
    """
    Draws a square-shaped blob with the given area (< 1) at
    the given coordinates.
    """
    hs = np.sqrt(area) / 2
    xcorners = array([x - hs, x + hs, x + hs, x - hs])
    ycorners = array([y - hs, y - hs, y + hs, y + hs])

    if ax is not None:
        handle = ax
    else:
        handle = plt

    return handle.fill(xcorners, ycorners, color=color_fn(w))


def _cb_labels(left_dims):
    """Creates plot labels for matrix elements in the computational basis.

    Parameters
    ----------
    left_dims : flat list of ints
        Dimensions of the left index of a density operator. E. g.
        [2, 3] for a qubit tensored with a qutrit.

    Returns
    -------
    left_labels, right_labels : lists of strings
        Labels for the left and right indices of a density operator
        (kets and bras, respectively).
    """
    # FIXME: assumes dims, such that we only need left_dims == dims[0].
    basis_labels = list(map(",".join, it.product(*[
        map(str, range(dim))
        for dim in left_dims
    ])))
    return [
        map(fmt.format, basis_labels) for fmt in
        (
            r"$\langle{}|$",
            r"$|{}\rangle$",
        )
    ]


# Adopted from the SciPy Cookbook.
def hinton(rho, x_basis=None, y_basis=None, color_style="scaled",
           label_top=True, *, cmap=None, colorbar=True, fig=None, ax=None):
    """Draws a Hinton diagram to visualize a density matrix or superoperator.

    Parameters
    ----------
    rho : qobj
        Input density matrix or superoperator.

        .. note::

            Hinton plots of superoperators are currently only
            supported for qubits.

    x_basis : list of strings, optional
        list of x ticklabels to represent x basis of the input.

    y_basis : list of strings, optional
        list of y ticklabels to represent y basis of the input.

    color_style : str, {"scaled", "threshold", "phase"}, default: "scaled"

        Determines how colors are assigned to each square:

        -  If set to ``"scaled"`` (default), each color is chosen by
           passing the absolute value of the corresponding matrix
           element into `cmap` with the sign of the real part.
        -  If set to ``"threshold"``, each square is plotted as
           the maximum of `cmap` for the positive real part and as
           the minimum for the negative part of the matrix element;
           note that this generalizes `"threshold"` to complex numbers.
        -  If set to ``"phase"``, each color is chosen according to
           the angle of the corresponding matrix element.

    label_top : bool, default: True
        If True, x ticklabels will be placed on top, otherwise
        they will appear below the plot.

    cmap : a matplotlib colormap instance, optional
        Color map to use when plotting.

    colorbar : bool, default: True
        Whether (True) or not (False) a colorbar should be attached.

    fig : a matplotlib Figure instance, optional
        The Figure canvas in which the plot will be drawn.

    ax : a matplotlib axes instance, optional
        The ax context in which the plot will be drawn.

    Returns
    -------
    fig, output : tuple
        A tuple of the matplotlib figure and the axes instance or animation
        instance used to produce the figure.

    Raises
    ------
    ValueError
        Input argument is not a quantum object.

    Examples
    --------
    >>> import qutip

    >>> dm = qutip.rand_dm(4)
    >>> fig, ax = qutip.hinton(dm)
    >>> fig.show()

    >>> qutip.settings.colorblind_safe = True
    >>> fig, ax = qutip.hinton(dm, color_style="threshold")
    >>> fig.show()
    >>> qutip.settings.colorblind_safe = False

    >>> fig, ax = qutip.hinton(dm, color_style="phase")
    >>> fig.show()
    """

    fig, ax = _is_fig_and_ax(fig, ax)

    if not isinstance(rho, list):
        rhos = [rho]
    else:
        rhos = rho

    _equal_shape(rhos)

    Ws = list()
    w_max = 0
    for rho in rhos:
        # Extract plotting data W from the input.
        if isinstance(rho, Qobj):
            if rho.isoper or rho.isoperket or rho.isoperbra:
                if rho.isoperket:
                    rho = vector_to_operator(rho)
                elif rho.isoperbra:
                    rho = vector_to_operator(rho.dag())
                W = rho.full()
                # Create default labels if none are given.
                labels = _cb_labels(rho.dims[0])
                if x_basis is None:
                    x_basis = list(labels[0])
                if y_basis is None:
                    y_basis = list(labels[1])

            elif rho.issuper:
                if not isqubitdims(rho.dims):
                    raise ValueError("Hinton plots of superoperators are "
                                     "currently only supported for qubits.")
                # Convert to a superoperator in the Pauli basis,
                # so that all the elements are real.
                sqobj = _to_superpauli(rho)
                nq = int(log2(sqobj.shape[0]) / 2)
                W = sqobj.full().T
                # Create default labels, too.
                labels = list(map("".join, it.product("IXYZ", repeat=nq)))
                if x_basis is None:
                    x_basis = labels
                if y_basis is None:
                    y_basis = labels

            else:
                raise ValueError(
                    "Input quantum object must be "
                    "an operator or superoperator.")
        else:
            W = rho
        Ws.append(W)

        height, width = W.shape

        w_max = max(1.25 * max(abs(np.array(W)).flatten()), w_max)
        if w_max <= 0.0:
            w_max = 1.0

    # Set color_fn here.
    if color_style == "scaled":
        if cmap is None:
            cmap = _diverging_cmap()

        def color_fn(w):
            w = np.abs(w) * np.sign(np.real(w))
            return cmap(int((w + w_max) * 256 / (2 * w_max)))
    elif color_style == "threshold":
        if cmap is None:
            cmap = _diverging_cmap()

        def color_fn(w):
            w = np.real(w)
            return cmap(255 if w > 0 else 0)
    elif color_style == "phase":
        if cmap is None:
            cmap = _cyclic_cmap()

        def color_fn(w):
            return cmap(int(255 * (np.angle(w) / 2 / np.pi + 0.5)))
    else:
        raise ValueError(
            "Unknown color style {} for Hinton diagrams.".format(color_style)
        )

    artist_list = list()
    ax.fill(array([0, width, width, 0]), array([0, 0, height, height]),
            color=cmap(128))
    for W in Ws:
        artist = list()
        for x in range(width):
            for y in range(height):
                _x = x + 1
                _y = y + 1
                artist += _blob(_x - 0.5, height - _y + 0.5, W[y, x],
                                w_max, min(1, abs(W[y, x]) / w_max),
                                color_fn=color_fn, ax=ax)
        artist_list.append(artist)

    if len(rhos) == 1:
        output = ax
    else:
        output = animation.ArtistAnimation(fig, artist_list, interval=50,
                                           blit=True, repeat_delay=1000)

    # axis
    if not (x_basis or y_basis):
        ax.axis('off')
    ax.axis('equal')
    ax.set_frame_on(False)

    # x axis
    xticks = 0.5 + np.arange(width)
    if x_basis:
        _set_ticklabels(ax, x_basis, xticks, 'x')
    if label_top:
        ax.xaxis.tick_top()

    # y axis
    yticks = 0.5 + np.arange(height)
    if y_basis:
        _set_ticklabels(ax, list(reversed(y_basis)), yticks, 'y')

    if colorbar:
        vmax = np.pi if color_style == "phase" else w_max
        norm = mpl.colors.Normalize(-vmax, vmax)
        cax, kw = mpl.colorbar.make_axes(ax, shrink=0.75, pad=.1)
        mpl.colorbar.ColorbarBase(cax, norm=norm, cmap=cmap)

    return fig, output


def sphereplot(values, theta, phi, *,
               cmap=None, colorbar=True, fig=None, ax=None):
    """Plots a matrix of values on a sphere

    Parameters
    ----------
    values : array
        Data set to be plotted

    theta : float
        Angle with respect to z-axis. Its range is between 0 and pi

    phi : float
        Angle in x-y plane. Its range is between 0 and 2*pi

    cmap : a matplotlib colormap instance, optional
        Color map to use when plotting.

    colorbar : bool, default: True
        Whether (True) or not (False) a colorbar should be attached.

    fig : a matplotlib Figure instance, optional
        The Figure canvas in which the plot will be drawn.

    ax : a matplotlib axes instance, optional
        The axes context in which the plot will be drawn.

    Returns
    -------
    fig, output : tuple
        A tuple of the matplotlib figure and the axes instance or animation
        instance used to produce the figure.
    """

    fig, ax = _is_fig_and_ax(fig, ax, projection='3d')

    if not isinstance(values, list):
        V = [values]
    else:
        V = values

    _equal_shape(V)

    r_and_ph = list()
    min_ph = pi
    max_ph = -pi
    for values in V:
        r = array(abs(values))
        ph = angle(values)
        min_ph = min(min_ph, ph.min())
        max_ph = max(max_ph, ph.max())
        r_and_ph.append((r, ph))

    # normalize color range based on phase angles in list ph
    norm = mpl.colors.Normalize(min_ph, max_ph)

    if cmap is None:
        cmap = _sequential_cmap()

    # plot with facecolors set to cm.jet colormap normalized to nrm
    thetam, phim = np.meshgrid(theta, phi)
    xx = sin(thetam) * cos(phim)
    yy = sin(thetam) * sin(phim)
    zz = cos(thetam)
    artist_list = list()
    for r, ph in r_and_ph:
        artist = [ax.plot_surface(r * xx, r * yy, r * zz, rstride=1, cstride=1,
                                  facecolors=cmap(norm(ph)), linewidth=0,)]
        artist_list.append(artist)

    if len(V) == 1:
        output = ax
    else:
        output = animation.ArtistAnimation(fig, artist_list, interval=50,
                                           blit=True, repeat_delay=1000)

    if colorbar:
        # create new axes on plot for colorbar and shrink it a bit.
        # pad shifts location of bar with repsect to the main plot
        cax, kw = mpl.colorbar.make_axes(ax, shrink=.66, pad=.05)

        # create new colorbar in axes cax with cmap and normalized to nrm like
        # our facecolors
        cb1 = mpl.colorbar.ColorbarBase(cax, cmap=cmap, norm=norm)
        # add our colorbar label
        cb1.set_label('Angle')

    return fig, output


def _remove_margins(axis):
    """
    removes margins about z = 0 and improves the style
    by monkey patching
    """
    def _get_coord_info_new(renderer):
        mins, maxs, centers, deltas, tc, highs = \
            _get_coord_info_old(renderer)
        mins += deltas / 4
        maxs -= deltas / 4
        return mins, maxs, centers, deltas, tc, highs

    _get_coord_info_old = axis._get_coord_info
    axis._get_coord_info = _get_coord_info_new


def _stick_to_planes(stick, azim, ax, M, spacing):
    """adjusts xlim and ylim in way that bars will
    stick to xz and yz planes
    """
    if stick is True:
        azim = azim % 360
        if 0 <= azim <= 90:
            ax.set_ylim(1 - .5,)
            ax.set_xlim(1 - .5,)
        elif 90 < azim <= 180:
            ax.set_ylim(1 - .5,)
            ax.set_xlim(0, M.shape[0] + (.5 - spacing))
        elif 180 < azim <= 270:
            ax.set_ylim(0, M.shape[1] + (.5 - spacing))
            ax.set_xlim(0, M.shape[0] + (.5 - spacing))
        elif 270 < azim < 360:
            ax.set_ylim(0, M.shape[1] + (.5 - spacing))
            ax.set_xlim(1 - .5,)


def _update_yaxis(spacing, M, ax, ylabels):
    """
    updates the y-axis
    """
    ytics = [y + (1 - (spacing / 2)) for y in range(M.shape[1])]
    ax.yaxis.set_major_locator(plt.FixedLocator(ytics))
    if ylabels:
        nylabels = len(ylabels)
        if nylabels != len(ytics):
            raise ValueError(f"got {nylabels} ylabels but needed {len(ytics)}")
        ax.set_yticklabels(ylabels)
    else:
        ax.set_yticklabels([str(y + 1) for y in range(M.shape[1])])
        ax.set_yticklabels([str(i) for i in range(M.shape[1])])
    ax.tick_params(axis='y', labelsize=14)
    ax.set_yticks([y + (1 - (spacing / 2)) for y in range(M.shape[1])])


def _update_xaxis(spacing, M, ax, xlabels):
    """
    updates the x-axis
    """
    xtics = [x + (1 - (spacing / 2)) for x in range(M.shape[0])]
    ax.xaxis.set_major_locator(plt.FixedLocator(xtics))
    if xlabels:
        nxlabels = len(xlabels)
        if nxlabels != len(xtics):
            raise ValueError(f"got {nxlabels} xlabels but needed {len(xtics)}")
        ax.set_xticklabels(xlabels)
    else:
        ax.set_xticklabels([str(x + 1) for x in range(M.shape[0])])
        ax.set_xticklabels([str(i) for i in range(M.shape[0])])
    ax.tick_params(axis='x', labelsize=14)
    ax.set_xticks([x + (1 - (spacing / 2)) for x in range(M.shape[0])])


def _update_zaxis(ax, z_min, z_max, zticks):
    """
    updates the z-axis
    """
    ax.zaxis.set_major_locator(plt.IndexLocator(1, 0.5))
    if isinstance(zticks, list):
        ax.set_zticks(zticks)
    ax.set_zlim3d([min(z_min, 0), z_max])


def _get_matrix_components(option, M, argument):
    if option == 'real':
        return np.real(M.flatten())
    elif option == 'img':
        return np.imag(M.flatten())
    elif option == 'abs':
        return np.abs(M.flatten())
    elif option == 'phase':
        return angle(M.flatten())
    else:
        raise ValueError("got an unexpected argument, "
                         f"{option} for {argument}")


def matrix_histogram(M, x_basis=None, y_basis=None, limits=None,
                     bar_style='real', color_limits=None, color_style='real',
                     options=None, *, cmap=None, colorbar=True,
                     fig=None, ax=None):
    """
    Draw a histogram for the matrix M, with the given x and y labels and title.

    Parameters
    ----------
    M : Matrix of Qobj
        The matrix to visualize

    x_basis : list of strings, optional
        list of x ticklabels

    y_basis : list of strings, optional
        list of y ticklabels

    limits : list/array with two float numbers, optional
        The z-axis limits [min, max]

    bar_style : str, {"real", "img", "abs", "phase"}, default: "real"

        -  If set to ``"real"`` (default), each bar is plotted
           as the real part of the corresponding matrix element
        -  If set to ``"img"``, each bar is plotted
           as the imaginary part of the corresponding matrix element
        -  If set to ``"abs"``, each bar is plotted
           as the absolute value of the corresponding matrix element
        -  If set to ``"phase"`` (default), each bar is plotted
           as the angle of the corresponding matrix element

    color_limits : list/array with two float numbers, optional
        The limits of colorbar [min, max]

    color_style : str, {"real", "img", "abs", "phase"}, default: "real"
        Determines how colors are assigned to each square:

        -  If set to ``"real"`` (default), each color is chosen
           according to the real part of the corresponding matrix element.
        -  If set to ``"img"``, each color is chosen according to
           the imaginary part of the corresponding matrix element.
        -  If set to ``"abs"``, each color is chosen according to
           the absolute value of the corresponding matrix element.
        -  If set to ``"phase"``, each color is chosen according to
           the angle of the corresponding matrix element.

    cmap : a matplotlib colormap instance, optional
        Color map to use when plotting.

    colorbar : bool, default: True
        show colorbar

    fig : a matplotlib Figure instance, optional
        The Figure canvas in which the plot will be drawn.

    ax : a matplotlib axes instance, optional
        The axes context in which the plot will be drawn.

    options : dict, optional
        A dictionary containing extra options for the plot.
        The names (keys) and values of the options are
        described below:

        'zticks' : list of numbers, optional
            A list of z-axis tick locations.

        'bars_spacing' : float, default: 0.1
            spacing between bars.

        'bars_alpha' : float, default: 1.
            transparency of bars, should be in range 0 - 1

        'bars_lw' : float, default: 0.5
            linewidth of bars' edges.

        'bars_edgecolor' : color, default: 'k'
            The colors of the bars' edges.
            Examples: 'k', (0.1, 0.2, 0.5) or '#0f0f0f80'.

        'shade' : bool, default: True
            Whether to shade the dark sides of the bars (True) or not (False).
            The shading is relative to plot's source of light.

        'azim' : float, default: -35
            The azimuthal viewing angle.

        'elev' : float, default: 35
            The elevation viewing angle.

        'stick' : bool, default: False
            Changes xlim and ylim in such a way that bars next to
            XZ and YZ planes will stick to those planes.
            This option has no effect if ``ax`` is passed as a parameter.

        'cbar_pad' : float, default: 0.04
            The fraction of the original axes between the colorbar
            and the new image axes.
            (i.e. the padding between the 3D figure and the colorbar).

        'cbar_to_z' : bool, default: False
            Whether to set the color of maximum and minimum z-values to the
            maximum and minimum colors in the colorbar (True) or not (False).

        'threshold': float, optional
            Threshold for when bars of smaller height should be transparent. If
            not set, all bars are colored according to the color map.

    Returns
    -------
    fig, output : tuple
        A tuple of the matplotlib figure and the axes instance or animation
        instance used to produce the figure.

    Raises
    ------
    ValueError
        Input argument is not valid.

    """

    # default options
    default_opts = {'zticks': None, 'bars_spacing': 0.2,
                    'bars_alpha': 1., 'bars_lw': 0.5, 'bars_edgecolor': 'k',
                    'shade': True, 'azim': -35, 'elev': 35, 'stick': False,
                    'cbar_pad': 0.04, 'cbar_to_z': False, 'threshold': None}

    # update default_opts from input options
    if options is None:
        options = dict()

    if isinstance(options, dict):
        # check if keys in options dict are valid
        if set(options) - set(default_opts):
            raise ValueError("invalid key(s) found in options: "
                             f"{', '.join(set(options) - set(default_opts))}")
        else:
            # updating default options
            default_opts.update(options)
            options = default_opts
    else:
        raise ValueError("options must be a dictionary")

    fig, ax = _is_fig_and_ax(fig, ax, projection='3d')

    if not isinstance(M, list):
        Ms = [M]
    else:
        Ms = M

    _equal_shape(Ms)

    for i in range(len(Ms)):
        M = Ms[i]
        if isinstance(M, Qobj):
            if x_basis is None:
                x_basis = list(_cb_labels([M.shape[0]])[0])
            if y_basis is None:
                y_basis = list(_cb_labels([M.shape[1]])[1])
            # extract matrix data from Qobj
            M = M.full()

        bar_M = _get_matrix_components(bar_style, M, 'bar_style')

        if isinstance(limits, list) and \
                len(limits) == 2:
            z_min = limits[0]
            z_max = limits[1]
        else:
            z_min = min(bar_M) if i == 0 else min(min(bar_M), z_min)
            z_max = max(bar_M) if i == 0 else max(max(bar_M), z_max)

            if z_min == z_max:
                z_min -= 0.1
                z_max += 0.1

        color_M = _get_matrix_components(color_style, M, 'color_style')

        if isinstance(color_limits, list) and \
                len(color_limits) == 2:
            c_min = color_limits[0]
            c_max = color_limits[1]
        else:
            if color_style == 'phase':
                c_min = -pi
                c_max = pi
            else:
                c_min = min(color_M) if i == 0 else min(min(color_M), c_min)
                c_max = min(color_M) if i == 0 else max(max(color_M), c_max)

            if c_min == c_max:
                c_min -= 0.1
                c_max += 0.1

    norm = mpl.colors.Normalize(c_min, c_max)

    if cmap is None:
        # change later
        if color_style == 'phase':
            cmap = _cyclic_cmap()
        else:
            cmap = _sequential_cmap()

    artist_list = list()
    for M in Ms:

        if isinstance(M, Qobj):
            M = M.full()

        bar_M = _get_matrix_components(bar_style, M, 'bar_style')
        color_M = _get_matrix_components(color_style, M, 'color_style')

        n = np.size(M)
        xpos, ypos = np.meshgrid(range(M.shape[0]), range(M.shape[1]))
        xpos = xpos.T.flatten() + 0.5
        ypos = ypos.T.flatten() + 0.5
        zpos = np.zeros(n)
        dx = dy = (1 - options['bars_spacing']) * np.ones(n)
        colors = cmap(norm(color_M))

        colors[:, 3] = options['bars_alpha']

        if options['threshold'] is not None:
            colors[:, 3] *= 1 * (bar_M >= options['threshold'])

            idx, = np.where(bar_M < options['threshold'])
            bar_M[idx] = 0

        artist = ax.bar3d(xpos, ypos, zpos, dx, dy, bar_M, color=colors,
                          edgecolors=options['bars_edgecolor'],
                          linewidths=options['bars_lw'],
                          shade=options['shade'])
        artist_list.append([artist])

    if len(Ms) == 1:
        output = ax
    else:
        output = animation.ArtistAnimation(fig, artist_list, interval=50,
                                           blit=True, repeat_delay=1000)

    # remove vertical lines on xz and yz plane
    ax.yaxis._axinfo["grid"]['linewidth'] = 0
    ax.xaxis._axinfo["grid"]['linewidth'] = 0

    # x axis
    _update_xaxis(options['bars_spacing'], M, ax, x_basis)

    # y axis
    _update_yaxis(options['bars_spacing'], M, ax, y_basis)

    # z axis
    _update_zaxis(ax, z_min, z_max, options['zticks'])

    # stick to xz and yz plane
    _stick_to_planes(options['stick'],
                     options['azim'], ax, M,
                     options['bars_spacing'])
    ax.view_init(azim=options['azim'], elev=options['elev'])

    # removing margins
    _remove_margins(ax.xaxis)
    _remove_margins(ax.yaxis)
    _remove_margins(ax.zaxis)

    # color axis
    if colorbar:
        cax, kw = mpl.colorbar.make_axes(ax, shrink=.75,
                                         pad=options['cbar_pad'])
        cb = mpl.colorbar.ColorbarBase(cax, cmap=cmap, norm=norm)

        if color_style == 'real':
            cb.set_label('real')
        elif color_style == 'img':
            cb.set_label('imaginary')
        elif color_style == 'abs':
            cb.set_label('absolute')
        else:
            cb.set_label('arg')
            if color_limits is None:
                cb.set_ticks([-pi, -pi / 2, 0, pi / 2, pi])
                cb.set_ticklabels(
                    (r'$-\pi$', r'$-\pi/2$', r'$0$', r'$\pi/2$', r'$\pi$'))

    return fig, output


def plot_energy_levels(H_list, h_labels=None, energy_levels=None, N=0, *,
                       fig=None, ax=None):
    """
    Plot the energy level diagrams for a list of Hamiltonians. Include
    up to N energy levels. For each element in H_list, the energy
    levels diagram for the cummulative Hamiltonian sum(H_list[0:n]) is plotted,
    where n is the index of an element in H_list.

    Parameters
    ----------

        H_list : List of Qobj
            A list of Hamiltonians.

        h_lables : List of string, optional
            A list of xticklabels for each Hamiltonian

        energy_levels : List of string, optional
            A list of  yticklabels to the left of energy levels of the initial
            Hamiltonian.

        N : int, default: 0
            The number of energy levels to plot

        fig : a matplotlib Figure instance, optional
            The Figure canvas in which the plot will be drawn.

        ax : a matplotlib axes instance, optional
            The axes context in which the plot will be drawn.

    Returns
    -------
    fig, ax : tuple
        A tuple of the matplotlib figure and axes instances used to produce
        the figure.

    Raises
    ------

        ValueError
            Input argument is not valid.

    """

    if not isinstance(H_list, list):
        raise ValueError("H_list must be a list of Qobj instances")

    fig, ax = _is_fig_and_ax(fig, ax)

    H = H_list[0]
    N = H.shape[0] if N == 0 else min(H.shape[0], N)
    xticks = []
    yticks = []
    x = 0
    evals0 = H.eigenenergies(eigvals=N)
    for e_idx, e in enumerate(evals0[:N]):
        ax.plot([x, x + 2], np.array([1, 1]) * e, 'b', linewidth=2)
        yticks.append(e)
    xticks.append(x + 1)
    x += 2

    for H1 in H_list[1:]:

        H = H + H1
        evals1 = H.eigenenergies()

        for e_idx, e in enumerate(evals1[:N]):
            ax.plot([x, x + 1], np.array([evals0[e_idx], e]), 'k:')
        x += 1

        for e_idx, e in enumerate(evals1[:N]):
            ax.plot([x, x + 2], np.array([1, 1]) * e, 'b', linewidth=2)
        xticks.append(x + 1)
        x += 2

        evals0 = evals1

    ax.set_frame_on(False)

    if energy_levels:
        yticks = np.unique(np.around(yticks, 1))
        _set_ticklabels(ax, energy_levels, yticks, 'y')
    else:
        # show eigenenergies
        yticks = np.unique(np.around(yticks, 1))
        ax.set_yticks(yticks)

    if h_labels:
        ax.get_xaxis().tick_bottom()
        _set_ticklabels(ax, h_labels, xticks, 'x')
    else:
        # hide xtick
        ax.tick_params(axis='x', which='both',
                       bottom=False, labelbottom=False)

    return fig, ax


def plot_fock_distribution(rho, fock_numbers=None, color="green",
                           unit_y_range=True, *, fig=None, ax=None):
    """
    Plot the Fock distribution for a density matrix (or ket) that describes
    an oscillator mode.

    Parameters
    ----------
    rho : :obj:`.Qobj`
        The density matrix (or ket) of the state to visualize.

    fock_numbers : list of strings, optional
        list of x ticklabels to represent fock numbers

    color : color or list of colors, default: "green"
        The colors of the bar faces.

    unit_y_range : bool, default: True
        Set y-axis limits [0, 1] or not

    fig : a matplotlib Figure instance, optional
        The Figure canvas in which the plot will be drawn.

    ax : a matplotlib axes instance, optional
        The axes context in which the plot will be drawn.

    Returns
    -------
    fig, output : tuple
        A tuple of the matplotlib figure and the axes instance or animation
        instance used to produce the figure.
    """

    fig, ax = _is_fig_and_ax(fig, ax)

    if not isinstance(rho, list):
        rhos = [rho]
    else:
        rhos = rho

    _equal_shape(rhos)

    artist_list = list()
    for rho in rhos:
        if isket(rho):
            rho = ket2dm(rho)

        N = rho.shape[0]

        artist = ax.bar(np.arange(N), np.real(rho.diag()),
                        color=color, alpha=0.6, width=0.8).patches
        artist_list.append(artist)

    if len(rhos) == 1:
        output = ax
    else:
        output = animation.ArtistAnimation(fig, artist_list, interval=50,
                                           blit=True, repeat_delay=1000)

    if fock_numbers:
        _set_ticklabels(ax, fock_numbers, np.arange(N), 'x', fontsize=12)

    if unit_y_range:
        ax.set_ylim(0, 1)

    ax.set_xlim(-.5, N)
    ax.set_xlabel('Fock number', fontsize=12)
    ax.set_ylabel('Occupation probability', fontsize=12)

    return fig, output


def plot_wigner(rho, xvec=None, yvec=None, method='clenshaw', projection='2d',
                g=sqrt(2), sparse=False, parfor=False, *,
                cmap=None, colorbar=False, fig=None, ax=None):
    """
    Plot the the Wigner function for a density matrix (or ket) that describes
    an oscillator mode.

    Parameters
    ----------
    rho : :obj:`.Qobj`
        The density matrix (or ket) of the state to visualize.

    xvec : array_like, optional
        x-coordinates at which to calculate the Wigner function.

    yvec : array_like, optional
        y-coordinates at which to calculate the Wigner function.  Does not
        apply to the 'fft' method.

    method : str {'clenshaw', 'iterative', 'laguerre', 'fft'}, default: 'clenshaw'
        The method used for calculating the wigner function. See the
        documentation for qutip.wigner for details.

    projection: str {'2d', '3d'}, default: '2d'
        Specify whether the Wigner function is to be plotted as a
        contour graph ('2d') or surface plot ('3d').

    g : float
        Scaling factor for `a = 0.5 * g * (x + iy)`, default `g = sqrt(2)`.
        See the documentation for qutip.wigner for details.

    sparse : bool {False, True}
        Flag for sparse format.
        See the documentation for qutip.wigner for details.

    parfor : bool {False, True}
        Flag for parallel calculation.
        See the documentation for qutip.wigner for details.

    cmap : a matplotlib cmap instance, optional
        The colormap.

    colorbar : bool, default: False
        Whether (True) or not (False) a colorbar should be attached to the
        Wigner function graph.

    fig : a matplotlib Figure instance, optional
        The Figure canvas in which the plot will be drawn.

    ax : a matplotlib axes instance, optional
        The axes context in which the plot will be drawn.

    Returns
    -------
    fig, output : tuple
        A tuple of the matplotlib figure and the axes instance or animation
        instance used to produce the figure.
    """

    if projection not in ('2d', '3d'):
        raise ValueError('Unexpected value of projection keyword argument')

    fig, ax = _is_fig_and_ax(fig, ax, projection)

    if not isinstance(rho, list):
        rhos = [rho]
    else:
        rhos = rho

    _equal_shape(rhos)

    wlim = 0
    Ws = list()
    xvec = np.linspace(-7.5, 7.5, 200) if xvec is None else xvec
    yvec = np.linspace(-7.5, 7.5, 200) if yvec is None else yvec
    for rho in rhos:
        if isket(rho):
            rho = ket2dm(rho)

        W0 = wigner(
            rho, xvec, yvec, method=method,
            g=g, sparse=sparse, parfor=parfor
        )

        W, yvec = W0 if isinstance(W0, tuple) else (W0, yvec)
        Ws.append(W)

        wlim = max(abs(W).max(), wlim)

    norm = mpl.colors.Normalize(-wlim, wlim)

    if cmap is None:
        cmap = _diverging_cmap()

    artist_list = list()
    for W in Ws:
        if projection == '2d':
            if parse_version(mpl.__version__) >= parse_version('3.8'):
                cf = [ax.contourf(xvec, yvec, W, 100, norm=norm, cmap=cmap)]
            else:
                cf = ax.contourf(xvec, yvec, W, 100, norm=norm,
                                 cmap=cmap).collections
        else:
            X, Y = np.meshgrid(xvec, yvec)
            cf = [ax.plot_surface(X, Y, W, rstride=5, cstride=5, linewidth=0.5,
                                  norm=norm, cmap=cmap)]
        artist_list.append(cf)

    if len(rhos) == 1:
        output = ax
    else:
        output = animation.ArtistAnimation(fig, artist_list, interval=50,
                                           blit=True, repeat_delay=1000)

    ax.set_xlabel(r'$\rm{Re}(\alpha)$', fontsize=12)
    ax.set_ylabel(r'$\rm{Im}(\alpha)$', fontsize=12)

    if colorbar:
        if projection == '2d':
            shrink = 1
        else:
            shrink = .75
        cax, kw = mpl.colorbar.make_axes(ax, shrink=shrink, pad=.1)
        cbar = mpl.colorbar.ColorbarBase(cax, cmap=cmap, norm=norm)

    return fig, output


def plot_expectation_values(results, ylabels=None, *,
                            fig=None, axes=None):
    """
    Visualize the results (expectation values) for an evolution solver.
    `results` is assumed to be an instance of Result, or a list of Result
    instances.

    Parameters
    ----------
    results : (list of) :class:`.Result`
        List of results objects returned by any of the QuTiP evolution solvers.

    ylabels : list of strings, optional
        The y-axis labels. List should be of the same length as `results`.

    fig : a matplotlib Figure instance, optional
        The Figure canvas in which the plot will be drawn.

    axes : (list of)  axes instances, optional
        The axes context in which the plot will be drawn.

    Returns
    -------
    fig, axes : tuple
        A tuple of the matplotlib figure and array of axes instances
        used to produce the figure.
    """
    if not isinstance(results, list):
        results = [results]

    n_e_ops = max([len(result.expect) for result in results])

    if axes is None:
        if fig is None:
            fig = plt.figure()
        axes = np.array([fig.add_subplot(n_e_ops, 1, i+1)
                         for i in range(n_e_ops)])

    # create np.ndarray if axes is one axes object or list
    if not isinstance(axes, np.ndarray):
        if not isinstance(axes, list):
            axes = [axes]
        axes = np.array(axes)

    for _, result in enumerate(results):
        for e_idx, e in enumerate(result.expect):
            axes[e_idx].plot(result.times, e,
                             label="%s [%d]" % (result.solver, e_idx))

    axes[n_e_ops - 1].set_xlabel("time", fontsize=12)
    for n in range(n_e_ops):
        if ylabels:
            axes[n].set_ylabel(ylabels[n], fontsize=12)

    return fig, axes


def plot_spin_distribution(P, THETA, PHI, projection='2d', *,
                           cmap=None, colorbar=False, fig=None, ax=None):
    """
    Plots a spin distribution (given as meshgrid data).

    Parameters
    ----------
    P : matrix
        Distribution values as a meshgrid matrix.

    THETA : matrix
        Meshgrid matrix for the theta coordinate. Its range is between 0 and pi

    PHI : matrix
        Meshgrid matrix for the phi coordinate. Its range is between 0 and 2*pi

    projection: str {'2d', '3d'}, default: '2d'
        Specify whether the spin distribution function is to be plotted as a 2D
        projection where the surface of the unit sphere is mapped on
        the unit disk ('2d') or surface plot ('3d').

    cmap : a matplotlib cmap instance, optional
        The colormap.

    colorbar : bool, default: False
        Whether (True) or not (False) a colorbar should be attached to the
        Wigner function graph.

    fig : a matplotlib figure instance, optional
        The figure canvas on which the plot will be drawn.

    ax : a matplotlib axis instance, optional
        The axis context in which the plot will be drawn.

    Returns
    -------
    fig, output : tuple
        A tuple of the matplotlib figure and the axes instance or animation
        instance used to produce the figure.
    """

    if projection in ('2d', '3d'):
        fig, ax = _is_fig_and_ax(fig, ax, projection)
    else:
        raise ValueError('Unexpected value of projection keyword argument')

    if not isinstance(P, list):
        Ps = [P]
    else:
        Ps = P

    _equal_shape(Ps)

    min_P = Ps[0].min()
    max_P = Ps[0].max()
    for P in Ps:
        min_P = min(min_P, P.min())
        max_P = max(max_P, P.max())

    if cmap is None:
        if min_P < -1e12:
            cmap = _diverging_cmap()
            norm = mpl.colors.Normalize(-max_P, max_P)
        else:
            cmap = _sequential_cmap()
            norm = mpl.colors.Normalize(min_P, max_P)

    artist_list = list()
    if projection == '2d':
        Y = (THETA - pi / 2) / (pi / 2)
        X = (pi - PHI) / pi * np.sqrt(cos(THETA - pi / 2))
        for P in Ps:
            artist_list.append([ax.pcolor(X, Y, P.real, cmap=cmap)])
        ax.set_xlabel(r'$\varphi$', fontsize=18)
        ax.set_ylabel(r'$\theta$', fontsize=18)
        ax.axis('equal')
        ax.set_xticks([-1, 0, 1])
        ax.set_xticklabels([r'$0$', r'$\pi$', r'$2\pi$'], fontsize=18)
        ax.set_yticks([-1, 0, 1])
        ax.set_yticklabels([r'$\pi$', r'$\pi/2$', r'$0$'], fontsize=18)
    else:
        xx = sin(THETA) * cos(PHI)
        yy = sin(THETA) * sin(PHI)
        zz = cos(THETA)
        for P in Ps:
            artist = [ax.plot_surface(xx, yy, zz, rstride=1, cstride=1,
                      facecolors=cmap(norm(P)), linewidth=0)]
            artist_list.append(artist)
        ax.view_init(azim=-35, elev=35)

    if len(Ps) == 1:
        output = ax
    else:
        output = animation.ArtistAnimation(fig, artist_list, interval=50,
                                           blit=True, repeat_delay=1000)

    if colorbar:
        cax, _ = mpl.colorbar.make_axes(ax, shrink=.66, pad=.1)
        cb1 = mpl.colorbar.ColorbarBase(cax, cmap=cmap, norm=norm)
        cb1.set_label('magnitude')

    return fig, output


#
# Qubism and other qubistic visualizations
#
def complex_array_to_rgb(X, theme='light', rmax=None):
    """
    Makes an array of complex number and converts it to an array of [r, g, b],
    where phase gives hue and saturation/value are given by the absolute value.
    Especially for use with imshow for complex plots.

    For more info on coloring, see:
        Emilia Petrisor,
        Visualizing complex-valued functions with Matplotlib and Mayavi
        https://nbviewer.ipython.org/github/empet/Math/blob/master/DomainColoring.ipynb

    Parameters
    ----------
    X : array
        Array (of any dimension) of complex numbers.

    theme : str {'light', 'dark'}, default: 'light'
        Set coloring theme for mapping complex values into colors.

    rmax : float, optional
        Maximal abs value for color normalization.
        If None (default), uses np.abs(X).max().

    Returns
    -------
    Y : array
        Array of colors (of shape X.shape + (3,)).

    """

    absmax = rmax or np.abs(X).max()
    if absmax == 0.:
        absmax = 1.
    Y = np.zeros(X.shape + (3,), dtype='float')
    Y[..., 0] = np.angle(X) / (2 * pi) % 1
    if theme == 'light':
        Y[..., 1] = np.clip(np.abs(X) / absmax, 0, 1)
        Y[..., 2] = 1
    elif theme == 'dark':
        Y[..., 1] = 1
        Y[..., 2] = np.clip(np.abs(X) / absmax, 0, 1)
    Y = mpl.colors.hsv_to_rgb(Y)
    return Y


def _index_to_sequence(i, dim_list):
    """
    For a matrix entry with index i it returns state it corresponds to.
    In particular, for dim_list=[2]*n it returns i written as a binary number.

    Parameters
    ----------
    i : int
        Index in a matrix.

    dim_list : list of int
        List of dimensions of consecutive particles.

    Returns
    -------
    seq : list
        List of coordinates for each particle.

    """
    res = []
    j = i
    for d in reversed(dim_list):
        j, s = divmod(j, d)
        res.append(s)
    return list(reversed(res))


def _sequence_to_index(seq, dim_list):
    """
    Inverse of _index_to_sequence.

    Parameters
    ----------
    seq : list of ints
        List of coordinates for each particle.

    dim_list : list of int
        List of dimensions of consecutive particles.

    Returns
    -------
    i : list
        Index in a matrix.

    """
    i = 0
    for s, d in zip(seq, dim_list):
        i *= d
        i += s

    return i


def _to_qubism_index_pair(i, dim_list, how='pairs'):
    """
    For a matrix entry with index i
    it returns x, y coordinates in qubism mapping.

    Parameters
    ----------
    i : int
        Index in a matrix.

    dim_list : list of int
        List of dimensions of consecutive particles.

    how : 'pairs' ('default'), 'pairs_skewed' or 'before_after'
        Type of qubistic plot.

    Returns
    -------
    x, y : tuple of ints
        List of coordinates for each particle.

    """
    seq = _index_to_sequence(i, dim_list)

    if how == 'pairs':
        y = _sequence_to_index(seq[::2], dim_list[::2])
        x = _sequence_to_index(seq[1::2], dim_list[1::2])
    elif how == 'pairs_skewed':
        dim_list2 = dim_list[::2]
        y = _sequence_to_index(seq[::2], dim_list2)
        seq2 = [(b - a) % d for a, b, d in zip(seq[::2], seq[1::2], dim_list2)]
        x = _sequence_to_index(seq2, dim_list2)
    elif how == 'before_after':
        # https://en.wikipedia.org/wiki/File:Ising-tartan.png
        n = len(dim_list)
        y = _sequence_to_index(reversed(seq[:(n // 2)]),
                               reversed(dim_list[:(n // 2)]))
        x = _sequence_to_index(seq[(n // 2):], dim_list[(n // 2):])
    else:
        raise Exception("No such 'how'.")

    return x, y


def _sequence_to_latex(seq, style='ket'):
    """
    For a sequence of particle states generate LaTeX code.

    Parameters
    ----------
    seq : list of ints
        List of coordinates for each particle.

    style : 'ket' (default), 'bra' or 'bare'
        Style of LaTeX (i.e. |01> or <01| or 01, respectively).

    Returns
    -------
    latex : str
        LaTeX output.

    """
    if style == 'ket':
        latex = "$\\left|{0}\\right\\rangle$"
    elif style == 'bra':
        latex = "$\\left\\langle{0}\\right|$"
    elif style == 'bare':
        latex = "${0}$"
    else:
        raise Exception("No such style.")
    return latex.format("".join(map(str, seq)))


def plot_qubism(ket, theme='light', how='pairs', grid_iteration=1,
                legend_iteration=0, *, fig=None, ax=None):
    """
    Qubism plot for pure states of many qudits.  Works best for spin chains,
    especially with even number of particles of the same dimension.  Allows to
    see entanglement between first 2k particles and the rest.

    .. note::

        colorblind_safe does not apply because of its unique colormap

    Parameters
    ----------
    ket : Qobj
        Pure state for plotting.

    theme : str {'light', 'dark'}, default: 'light'
        Set coloring theme for mapping complex values into colors.
        See: complex_array_to_rgb.

    how : str {'pairs', 'pairs_skewed' or 'before_after'}, default: 'pairs'
        Type of Qubism plotting.  Options:

        - 'pairs' - typical coordinates,
        - 'pairs_skewed' - for ferromagnetic/antriferromagnetic plots,
        - 'before_after' - related to Schmidt plot (see also: plot_schmidt).

    grid_iteration : int, default: 1
        Helper lines to be drawn on plot.
        Show tiles for 2*grid_iteration particles vs all others.

    legend_iteration : int or 'grid_iteration' or 'all', default: 0
        Show labels for first ``2*legend_iteration`` particles.  Option
        'grid_iteration' sets the same number of particles as for
        grid_iteration.  Option 'all' makes label for all particles.  Typically
        it should be 0, 1, 2 or perhaps 3.

    fig : a matplotlib figure instance, optional
        The figure canvas on which the plot will be drawn.

    ax : a matplotlib axis instance, optional
        The axis context in which the plot will be drawn.

    Returns
    -------
    fig, output : tuple
        A tuple of the matplotlib figure and the axes instance or animation
        instance used to produce the figure.

    Notes
    -----
    See also [1]_.

    References
    ----------
    .. [1] J. Rodriguez-Laguna, P. Migdal, M. Ibanez Berganza, M. Lewenstein
       and G. Sierra, *Qubism: self-similar visualization of many-body
       wavefunctions*, `New J. Phys. 14 053028
       <https://dx.doi.org/10.1088/1367-2630/14/5/053028>`_, arXiv:1112.3560
       (2012), open access.
    """

    fig, ax = _is_fig_and_ax(fig, ax)

    if not isinstance(ket, list):
        kets = [ket]
    else:
        kets = ket

    _equal_shape(kets)

    artist_list = list()
    for ket in kets:
        if not isket(ket):
            raise Exception("Qubism works only for pure states, i.e. kets.")
            # add for dm? (perhaps a separate function, plot_qubism_dm)

        dim_list = ket.dims[0]
        n = len(dim_list)

        # for odd number of particles - pixels are rectangular
        if n % 2 == 1:
            ket = tensor(ket, Qobj([1] * dim_list[-1]))
            dim_list = ket.dims[0]
            n += 1

        ketdata = ket.full()

        if how == 'pairs':
            dim_list_y = dim_list[::2]
            dim_list_x = dim_list[1::2]
        elif how == 'pairs_skewed':
            dim_list_y = dim_list[::2]
            dim_list_x = dim_list[1::2]
            if dim_list_x != dim_list_y:
                raise Exception("For 'pairs_skewed' pairs " +
                                "of dimensions need to be the same.")
        elif how == 'before_after':
            dim_list_y = list(reversed(dim_list[:(n // 2)]))
            dim_list_x = dim_list[(n // 2):]
        else:
            raise Exception("No such 'how'.")

        size_x = np.prod(dim_list_x)
        size_y = np.prod(dim_list_y)

        qub = np.zeros([size_x, size_y], dtype=complex)
        for i in range(ketdata.size):
            qub[_to_qubism_index_pair(i, dim_list, how=how)] = ketdata[i, 0]
        qub = qub.transpose()

        artist = [ax.imshow(complex_array_to_rgb(qub, theme=theme),
                  interpolation="none",
                  extent=(0, size_x, 0, size_y))]
        artist_list.append(artist)

    if len(kets) == 1:
        output = ax
    else:
        output = animation.ArtistAnimation(fig, artist_list, interval=50,
                                           blit=True, repeat_delay=1000)

    quadrants_x = np.prod(dim_list_x[:grid_iteration])
    quadrants_y = np.prod(dim_list_y[:grid_iteration])

    ticks_x = [size_x // quadrants_x * i for i in range(1, quadrants_x)]
    ticks_y = [size_y // quadrants_y * i for i in range(1, quadrants_y)]

    ax.set_xticks(ticks_x)
    ax.set_xticklabels([""] * (quadrants_x - 1))
    ax.set_yticks(ticks_y)
    ax.set_yticklabels([""] * (quadrants_y - 1))
    theme2color_of_lines = {'light': '#000000',
                            'dark': '#FFFFFF'}
    ax.grid(True, color=theme2color_of_lines[theme])

    if legend_iteration == 'all':
        label_n = n // 2
    elif legend_iteration == 'grid_iteration':
        label_n = grid_iteration
    else:
        try:
            label_n = int(legend_iteration)
        except:
            raise Exception("No such option for legend_iteration keyword " +
                            "argument. Use 'all', 'grid_iteration' or an " +
                            "integer.")

    if label_n:

        if how == 'before_after':
            dim_list_small = list(reversed(dim_list_y[-label_n:])) \
                + dim_list_x[:label_n]
        else:
            dim_list_small = []
            for j in range(label_n):
                dim_list_small.append(dim_list_y[j])
                dim_list_small.append(dim_list_x[j])

        scale_x = float(size_x) / np.prod(dim_list_x[:label_n])
        shift_x = 0.5 * scale_x
        scale_y = float(size_y) / np.prod(dim_list_y[:label_n])
        shift_y = 0.5 * scale_y

        bbox = ax.get_window_extent().transformed(
            fig.dpi_scale_trans.inverted())
        fontsize = 35 * bbox.width / np.prod(dim_list_x[:label_n]) / label_n
        opts = {'fontsize': fontsize,
                'color': theme2color_of_lines[theme],
                'horizontalalignment': 'center',
                'verticalalignment': 'center'}
        for i in range(np.prod(dim_list_small)):
            x, y = _to_qubism_index_pair(i, dim_list_small, how=how)
            seq = _index_to_sequence(i, dim_list=dim_list_small)
            ax.text(scale_x * x + shift_x,
                    size_y - (scale_y * y + shift_y),
                    _sequence_to_latex(seq),
                    **opts)

    return fig, output


def plot_schmidt(ket, theme='light', splitting=None,
                 labels_iteration=(3, 2), *, fig=None, ax=None):
    """
    Plotting scheme related to Schmidt decomposition.
    Converts a state into a matrix (A_ij -> A_i^j),
    where rows are first particles and columns - last.

    See also: plot_qubism with how='before_after' for a similar plot.

    .. note::

        colorblind_safe does not apply because of its unique colormap

    Parameters
    ----------
    ket : Qobj
        Pure state for plotting.

    theme : str {'light', 'dark'}, default: 'light'
        Set coloring theme for mapping complex values into colors.
        See: complex_array_to_rgb.

    splitting : int, optional
        Plot for a number of first particles versus the rest.
        If not given, it is (number of particles + 1) // 2.

    labels_iteration : int or pair of ints, default: (3, 2)
        Number of particles to be shown as tick labels,
        for first (vertical) and last (horizontal) particles, respectively.

    fig : a matplotlib figure instance, optional
        The figure canvas on which the plot will be drawn.

    ax : a matplotlib axis instance, optional
        The axis context in which the plot will be drawn.

    Returns
    -------
    fig, output : tuple
        A tuple of the matplotlib figure and the axes instance or animation
        instance used to produce the figure.

    """

    fig, ax = _is_fig_and_ax(fig, ax)

    if not isinstance(ket, list):
        kets = [ket]
    else:
        kets = ket

    _equal_shape(kets)

    artist_list = list()

    for ket in kets:
        if not isket(ket):
            err = "Schmidt plot works only for pure states, i.e. kets."
            raise Exception(err)

        dim_list = ket.dims[0]

        if splitting is None:
            splitting = (len(dim_list) + 1) // 2

        if isinstance(labels_iteration, int):
            labels_iteration = labels_iteration, labels_iteration

        ketdata = ket.full()

        dim_list_y = dim_list[:splitting]
        dim_list_x = dim_list[splitting:]

        size_x = np.prod(dim_list_x)
        size_y = np.prod(dim_list_y)

        ketdata = ketdata.reshape((size_y, size_x))

        artist = [ax.imshow(complex_array_to_rgb(ketdata, theme=theme),
                            interpolation="none",
                            extent=(0, size_x, 0, size_y))]
        artist_list.append(artist)

    if len(kets) == 1:
        output = ax
    else:
        output = animation.ArtistAnimation(fig, artist_list, interval=50,
                                           blit=True, repeat_delay=1000)

    dim_list_small_x = dim_list_x[:labels_iteration[1]]
    dim_list_small_y = dim_list_y[:labels_iteration[0]]

    quadrants_x = np.prod(dim_list_small_x)
    quadrants_y = np.prod(dim_list_small_y)

    ticks_x = [size_x / quadrants_x * (i + 0.5)
               for i in range(quadrants_x)]
    ticks_y = [size_y / quadrants_y * (quadrants_y - i - 0.5)
               for i in range(quadrants_y)]

    labels_x = [_sequence_to_latex(_index_to_sequence(i*size_x // quadrants_x,
                                                      dim_list=dim_list_x))
                for i in range(quadrants_x)]
    labels_y = [_sequence_to_latex(_index_to_sequence(i*size_y // quadrants_y,
                                                      dim_list=dim_list_y))
                for i in range(quadrants_y)]

    ax.set_xticks(ticks_x)
    ax.set_xticklabels(labels_x)
    ax.set_yticks(ticks_y)
    ax.set_yticklabels(labels_y)
    ax.set_xlabel("last particles")
    ax.set_ylabel("first particles")

    return fig, output



#cython: language_level=3
from qutip.core.data cimport Data, Dense
from qutip.core.cy.qobjevo cimport QobjEvo

cdef class _StochasticSystem:
    cdef readonly list c_ops
    cdef readonly QobjEvo L
    cdef readonly int num_collapse
    cdef Data state
    cdef double t

    cpdef Data drift(self, t, Data state)

    cpdef list diffusion(self, t, Data state)

    cpdef void set_state(self, double t, Dense state) except *

    cpdef Data a(self)
    cpdef Data bi(self, int i)
    cpdef Data Libj(self, int i, int j)
    cpdef Data Lia(self, int i)
    cpdef Data L0bi(self, int i)
    cpdef Data LiLjbk(self, int i, int j, int k)
    cpdef Data L0a(self)



#cython: language_level=3
#cython: boundscheck=False, wraparound=False, initializedcheck=False, cdvision=True

import numpy as np
import numbers
import itertools
from functools import partial

import qutip
from .. import Qobj
from .. import data as _data
from ..dimensions import Dimensions
from ..coefficient import coefficient, CompilationOptions
from ._element import *
from qutip.settings import settings

from qutip.core.cy._element cimport _BaseElement
from qutip.core.data cimport Dense, Data, dense
from qutip.core.data.expect cimport *
from qutip.core.data.reshape cimport (column_stack_dense, column_unstack_dense)
from qutip.core.cy.coefficient cimport Coefficient
from libc.math cimport fabs

__all__ = ['QobjEvo']

cdef class QobjEvo:
    """
    A class for representing time-dependent quantum objects, such as quantum
    operators and states.

    Importantly, :obj:`.QobjEvo` instances are used to represent such
    time-dependent quantum objects when working with QuTiP solvers.

    A :obj:`.QobjEvo` instance may be constructed from one of the following:

      * a callable ``f(t: double, args: dict) -> Qobj`` that returns the
        value of the quantum object at time ``t``.

      * a ``[Qobj, Coefficient]`` pair, where the :obj:`Coefficient` may be any
        item that :func:`.coefficient` can accept (e.g. a function, a numpy
        array of coefficient values, a string expression).

      * a :obj:`.Qobj` (which creates a constant :obj:`.QobjEvo` term).

      * a list of such callables, pairs or :obj:`.Qobj`\s.

      * a :obj:`.QobjEvo` (in which case a copy is created, all other arguments
        are ignored except ``args`` which, if passed, replaces the existing
        arguments).

    Parameters
    ----------
    Q_object : callable, list or :obj:`.Qobj`
        A specification of the time-depedent quantum object. See the
        paragraph above for a full description and the examples section below
        for examples.

    args : dict, optional
        A dictionary that contains the arguments for the coefficients.
        Arguments may be omitted if no function or string coefficients that
        require arguments are present.

    tlist : array-like, optional
        A list of times corresponding to the values of the coefficients
        supplied as numpy arrays. If no coefficients are supplied as numpy
        arrays, ``tlist`` may be omitted, otherwise it is required.

        The times in ``tlist`` do not need to be equidistant, but must
        be sorted.

        By default, a cubic spline interpolation will be used to interpolate
        the value of the (numpy array) coefficients at time ``t``. If the
        coefficients are to be treated as step functions, pass the argument
        ``order=0`` (see below).

    order : int, default=3
        Order of the spline interpolation that is to be used to interpolate
        the value of the (numpy array) coefficients at time ``t``.
        ``0`` use previous or left value.

    copy : bool, default=True
        Whether to make a copy of the :obj:`.Qobj` instances supplied in
        the ``Q_object`` parameter.

    compress : bool, default=True
        Whether to compress the :obj:`.QobjEvo` instance terms after the
        instance has been created.

        This sums the constant terms in a single term and combines
        ``[Qobj, coefficient]`` pairs with the same :obj:`.Qobj` into a single
        pair containing the sum of the coefficients.

        See :meth:`compress`.

    function_style : {None, "pythonic", "dict", "auto"}
        The style of function signature used by callables in ``Q_object``.
        If style is ``None``, the value of
        ``qutip.settings.core["function_coefficient_style"]``
        is used. Otherwise the supplied value overrides the global setting.


    boundary_conditions : 2-Tuple, str or None, optional
        Boundary conditions for spline evaluation. Default value is `None`.
        Correspond to `bc_type` of scipy.interpolate.make_interp_spline.
        Refer to Scipy's documentation for further details:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.make_interp_spline.html

    Attributes
    ----------
    dims : list
        List of dimensions keeping track of the tensor structure.

    shape : (int, int)
        List of dimensions keeping track of the tensor structure.

    type : str
        Type of quantum object: 'bra', 'ket', 'oper', 'operator-ket',
        'operator-bra', or 'super'.

    superrep : str
        Representation used if `type` is 'super'. One of 'super'
        (Liouville form) or 'choi' (Choi matrix with tr = dimension).

    Examples
    --------

    A :obj:`.QobjEvo` constructed from a function:

    .. code-block::

        def f(t, args):
            return qutip.qeye(N) * np.exp(args['w'] * t)

        QobjEvo(f, args={'w': 1j})


    For list based :obj:`.QobjEvo`, the list must consist of :obj:`.Qobj` or
    ``[Qobj, Coefficient]`` pairs:

    .. code-block::

        QobjEvo([H0, [H1, coeff1], [H2, coeff2]], args=args)

    The coefficients may be specified either using a :obj:`Coefficient` object
    or by a function, string, numpy array or any object that can be passed to
    the :func:`.coefficient` function. See the documentation of
    :func:`.coefficient` for a full description.

    An example of a coefficient specified by a function:

    .. code-block::

        def f1_t(t, args):
            return np.exp(-1j * t * args["w1"])

        QobjEvo([[H1, f1_t]], args={"w1": 1.})

    And of coefficients specified by string expressions:

    .. code-block::

        H = QobjEvo(
            [H0, [H1, 'exp(-1j*w1*t)'], [H2, 'cos(w2*t)']],
            args={"w1": 1., "w2": 2.}
        )

    Coefficients maybe also be expressed as numpy arrays giving a list
    of the coefficient values:

    .. code-block:: python

        tlist = np.logspace(-5, 0, 100)
        H = QobjEvo(
            [H0, [H1, np.exp(-1j * tlist)], [H2, np.cos(2. * tlist)]],
            tlist=tlist
        )

    The coeffients array must have the same len as the tlist.

    A :obj:`.QobjEvo` may also be built using simple arithmetic operations
    combining :obj:`.Qobj` with :obj:`Coefficient`, for example:

    .. code-block:: python

        coeff = qutip.coefficient("exp(-1j*w1*t)", args={"w1": 1})
        qevo = H0 + H1 * coeff

    """
    def __init__(QobjEvo self, Q_object, args=None, *, copy=True, compress=True,
                 function_style=None,
                 tlist=None, order=3, boundary_conditions=None):
        if isinstance(Q_object, QobjEvo):
            self._dims = Q_object._dims
            self.shape = Q_object.shape
            self.elements = (<QobjEvo> Q_object).elements.copy()
            self._feedback_functions = Q_object._feedback_functions.copy()
            self._solver_only_feedback = Q_object._solver_only_feedback.copy()
            if args:
                self.arguments(args)
            if compress:
                self.compress()
            return

        self.elements = []
        self._dims = None
        self.shape = (0, 0)
        self._feedback_functions = {}
        self._solver_only_feedback = {}
        args = self._read_args(args or {})

        if (
            isinstance(Q_object, list)
            and len(Q_object) == 2
            and isinstance(Q_object[0], Qobj)
            and not isinstance(Q_object[1], (Qobj, list))
        ):
            # The format is [Qobj, coefficient]
            Q_object = [Q_object]

        if isinstance(Q_object, list):
            for op in Q_object:
                self.elements.append(
                    self._read_element(
                        op, copy=copy, tlist=tlist, args=args, order=order,
                        function_style=function_style,
                        boundary_conditions=boundary_conditions
                    )
                )
        else:
            self.elements.append(
                self._read_element(
                    Q_object, copy=copy, tlist=tlist, args=args, order=order,
                    function_style=function_style,
                    boundary_conditions=boundary_conditions
                )
            )

        for key in self._feedback_functions:
            # During _read_args, the dims could not have been set yet.
            # To set the dims, for function QobjEvo, they need to be called.
            # But to be called, the feedback args need to be read...
            self._feedback_functions[key].check_consistency(self._dims)

        if compress:
            self.compress()

    def __repr__(self):
        cls = self.__class__.__name__
        repr_str = f'{cls}: dims = {self.dims}, shape = {self.shape}, '
        repr_str += f'type = {self.type}, superrep = {self.superrep}, '
        repr_str += f'isconstant = {self.isconstant}, '
        repr_str += f'num_elements = {self.num_elements}'
        feedback_pairs = []
        for key, val in self._feedback_functions.items():
            feedback_pairs.append(key + ":" + repr(val))
        for key, val in self._solver_only_feedback.items():
            feedback_pairs.append(key + ":" + val)
        if feedback_pairs:
            repr_str += f', feedback = {feedback_pairs}'

        return repr_str

    def _read_element(self, op, copy, tlist, args, order, function_style,
                      boundary_conditions):
        """ Read a Q_object item and return an element for that item. """
        if isinstance(op, Qobj):
            out = _ConstantElement(op.copy() if copy else op)
            qobj = op
        elif isinstance(op, list):
            out = _EvoElement(
                op[0].copy() if copy else op[0],
                coefficient(op[1], tlist=tlist, args=args, order=order,
                            boundary_conditions=boundary_conditions)
            )
            qobj = op[0]
        elif isinstance(op, _BaseElement):
            out = op
            qobj = op.qobj(0)
        elif callable(op):
            out = _FuncElement(op, args, style=function_style)
            qobj = out.qobj(0)
            if not isinstance(qobj, Qobj):
                raise TypeError(
                    "Function based time-dependent elements must have the"
                    " signature f(t: double, args: dict) -> Qobj, but"
                    " {!r} returned: {!r}".format(op, qobj)
                )
        else:
            raise TypeError(
                "QobjEvo terms should be Qobjs, a list of [Qobj, coefficient],"
                " or a function f(t: double, args: dict) -> Qobj, but"
                " received: {!r}".format(op)
            )

        if self._dims is None:
            self._dims = qobj._dims
            self.shape = qobj.shape
        elif self._dims != qobj._dims:
            raise ValueError(
                f"QobjEvo term {op!r} has dims {qobj.dims!r} and shape"
                f" {qobj.shape!r} but previous terms had dims {self.dims!r}"
                f" and shape {self.shape!r}."
            )

        return out

    @classmethod
    def _restore(cls, elements, dims, shape):
        """Recreate a QobjEvo without using __init__. """
        cdef QobjEvo out = cls.__new__(cls)
        out.elements = elements
        out._dims = dims
        out.shape = shape
        return out

    def _getstate(self):
        """ Obtain the state """
        # For jax pytree representation
        # auto_pickle create similar method __getstate__, but since it's
        # automatically created, it could change depending on cython version
        # etc., so we create our own.
        return {
            "elements": self.elements,
            "dims": self._dims,
            "shape": self.shape,
        }

    def __call__(self, double t, dict _args=None, **kwargs):
        """
        Get the :obj:`.Qobj` at ``t``.

        Parameters
        ----------
        t : float
            Time at which the :obj:`.QobjEvo` is to be evalued.

        _args : dict [optional]
            New arguments as a dict. Update args with ``arguments(new_args)``.

        **kwargs :
            New arguments as a keywors. Update args with
            ``arguments(**new_args)``.

        Notes
        -----
        If both the positional ``_args`` and keywords are passed new values
        from both will be used. If a key is present with both, the
        ``_args`` dict value will take priority.
        """
        if _args is not None or kwargs:
            if _args is not None:
                kwargs.update(_args)
            return QobjEvo(self, args=kwargs)(t)

        t = self._prepare(t, None)

        if self.isconstant:
            # For constant QobjEvo's, we sum the contained Qobjs directly in
            # order to retain the cached values of attributes like .isherm when
            # possible, rather than calling _call(t) which may lose this cached
            # information.
            return sum(element.qobj(t) for element in self.elements)

        cdef _BaseElement part = self.elements[0]
        cdef double complex coeff = part.coeff(t)
        obj = part.qobj(t)
        cdef Data out = _data.mul(obj.data, coeff)
        cdef bint isherm = <bint> obj._isherm and coeff.imag == 0
        for element in self.elements[1:]:
            part = <_BaseElement> element
            coeff = part.coeff(t)
            obj = part.qobj(t)
            isherm &= <bint> obj._isherm and coeff.imag == 0
            out = _data.add(out, obj.data, coeff)

        return Qobj(out, dims=self._dims, copy=False, isherm=isherm or None)

    cpdef Data _call(QobjEvo self, double t):
        t = self._prepare(t, None)
        cdef Data out
        cdef _BaseElement part = self.elements[0]
        out = _data.mul(part.data(t),
                        part.coeff(t))
        for element in self.elements[1:]:
            part = <_BaseElement> element

            out = _data.add(
                out,
                part.data(t),
                part.coeff(t)
            )
        return out

    cdef object _prepare(QobjEvo self, object t, Data state=None):
        """ Precomputation before computing getting the element at `t`"""
        # We keep the function for feedback eventually
        if self._feedback_functions and state is not None:
            new_args = {
                key: func(t, state)
                for key, func in self._feedback_functions.items()
            }
            cache = []
            self.elements = [
                element.replace_arguments(new_args, cache=cache)
                for element in self.elements
            ]

        return t

    def copy(QobjEvo self):
        """Return a copy of this :obj:`.QobjEvo`"""
        return QobjEvo(self, compress=False)

    def arguments(QobjEvo self, dict _args=None, **kwargs):
        """
        Update the arguments.

        Parameters
        ----------
        _args : dict [optional]
            New arguments as a dict. Update args with ``arguments(new_args)``.

        **kwargs :
            New arguments as a keywors. Update args with
            ``arguments(**new_args)``.

        Notes
        -----
        If both the positional ``_args`` and keywords are passed new values
        from both will be used. If a key is present with both, the ``_args``
        dict value will take priority.
        """
        if _args is not None:
            kwargs.update(_args)
        cache = []
        kwargs = self._read_args(kwargs)
        self.elements = [
            element.replace_arguments(kwargs, cache=cache)
            for element in self.elements
        ]

    def _read_args(self, args):
        """
        Filter feedback args from normal args.
        """
        new_args = {}
        for key, val in args.items():
            if isinstance(val, _Feedback):
                new_args[key] = val.default
                if self._dims is not None:
                    val.check_consistency(self._dims)
                if callable(val):
                    self._feedback_functions[key] = val
                else:
                    self._solver_only_feedback[key] = val.code
            else:
                new_args[key] = val
                if key in self._feedback_functions:
                    del self._feedback_functions[key]
                if key in self._solver_only_feedback:
                    del self._solver_only_feedback[key]

        return new_args

    def _register_feedback(self, solvers_feeds, solver):
        """
        Receive feedback source from solver.

        Parameters
        ----------
        solvers_feeds : dict[str]
            When ``feedback={key: solver_specific}`` is used, update arguments
            with ``args[key] = solvers_feeds[solver_specific]``.

        solver: str
            Name of the solver for the error message.
        """
        new_args = {}
        for key, feed in self._solver_only_feedback.items():
            if feed not in solvers_feeds:
                raise ValueError(
                    f"Desired feedback {key} is not available for the {solver}."
                )
            new_args[key] = solvers_feeds[feed]
        self.arguments(**new_args)

    def _update_feedback(QobjEvo self, QobjEvo other=None):
        """
        Merge feedback from ``op`` into self.
        """
        if other is not None:
            if not self._feedback_functions and other._feedback_functions:
                self._feedback_functions = other._feedback_functions.copy()
            elif other._feedback_functions:
                self._feedback_functions.update(other._feedback_functions)

            if not self._solver_only_feedback and other._solver_only_feedback:
                self._solver_only_feedback = other._solver_only_feedback.copy()
            elif other._solver_only_feedback:
                self._solver_only_feedback.update(other._solver_only_feedback)

        if self._feedback_functions:
            for key in self._feedback_functions:
                self._feedback_functions[key].check_consistency(self._dims)

    ###########################################################################
    # Math function                                                           #
    ###########################################################################
    def __add__(left, right):
        if isinstance(left, QobjEvo):
            self = left
            other = right
        else:
            self = right
            other = left
        if not isinstance(other, (Qobj, QobjEvo, numbers.Number)):
            return NotImplemented
        res = self.copy()
        res += other
        return res

    def __radd__(self, other):
        if not isinstance(other, (Qobj, QobjEvo, numbers.Number)):
            return NotImplemented
        res = self.copy()
        res += other
        return res

    def __iadd__(QobjEvo self, other):
        cdef _BaseElement element
        if isinstance(other, QobjEvo):
            if other._dims != self._dims:
                raise TypeError("incompatible dimensions" +
                                str(self.dims) + ", " + str(other.dims))
            for element in (<QobjEvo> other).elements:
                self.elements.append(element)
            self._update_feedback(other)

        elif isinstance(other, Qobj):
            if other._dims != self._dims:
                raise TypeError("incompatible dimensions" +
                                str(self.dims) + ", " + str(other.dims))
            self.elements.append(_ConstantElement(other))
        elif (
            isinstance(other, numbers.Number) and
            self._dims[0] == self._dims[1]
        ):
            self.elements.append(_ConstantElement(other * qutip.qeye_like(self)))
        else:
            return NotImplemented
        return self

    def __sub__(left, right):
        if isinstance(left, QobjEvo):
            res = left.copy()
            res += -right
            return res
        else:
            res = -right.copy()
            res += left
            return res

    def __rsub__(self, other):
        if not isinstance(other, (Qobj, QobjEvo, numbers.Number)):
            return NotImplemented
        res = -self
        res += other
        return res

    def __isub__(self, other):
        if not isinstance(other, (Qobj, QobjEvo, numbers.Number)):
            return NotImplemented
        self += (-other)
        return self

    def __matmul__(left, right):
        cdef QobjEvo res
        if isinstance(left, QobjEvo):
            return left.copy().__imatmul__(right)
        elif isinstance(left, Qobj):
            if left._dims[1] != (<QobjEvo> right)._dims[0]:
                raise TypeError("incompatible dimensions" +
                                 str(left.dims[1]) + ", " +
                                 str((<QobjEvo> right).dims[0]))
            res = right.copy()
            res._dims = Dimensions(left._dims[0], right._dims[1])
            res.shape = (left.shape[0], right.shape[1])
            left = _ConstantElement(left)
            res.elements = [left @ element for element in res.elements]
            res._update_feedback()

            return res

        else:
            return NotImplemented

    def __rmatmul__(QobjEvo self, other):
        cdef QobjEvo res
        if isinstance(other, Qobj):
            if other._dims[1] != self._dims[0]:
                raise TypeError("incompatible dimensions" +
                                 str(other.dims[1]) + ", " +
                                 str(self.dims[0]))
            res = self.copy()
            res._dims = Dimensions([other._dims[0], res._dims[1]])
            res.shape = (other.shape[0], res.shape[1])
            other = _ConstantElement(other)
            res.elements = [other @ element for element in res.elements]
            res._update_feedback()
            return res
        else:
            return NotImplemented

    def __imatmul__(QobjEvo self, other):
        if isinstance(other, (Qobj, QobjEvo)):
            if self.dims[1] != other.dims[0]:
                raise TypeError("incompatible dimensions" +
                                str(self.dims[1]) + ", " +
                                str(other.dims[0]))
            self._dims = Dimensions([self.dims[0], other.dims[1]])
            self.shape = (self.shape[0], other.shape[1])
            if isinstance(other, Qobj):
                other = _ConstantElement(other)
                self.elements = [element @ other for element in self.elements]
                self._update_feedback()

            elif isinstance(other, QobjEvo):
                self.elements = [left @ right
                    for left, right in itertools.product(
                        self.elements, (<QobjEvo> other).elements
                    )]
                self._update_feedback(other)

        else:
            return NotImplemented
        return self

    def __mul__(left, right):
        if isinstance(left, QobjEvo):
            return left.copy().__imul__(right)
        elif isinstance(left, Qobj):
            return right.__rmatmul__(left)
        elif isinstance(left, (numbers.Number, Coefficient)):
            return right.copy().__imul__(left)
        else:
            return NotImplemented

    def __rmul__(self, other):
        if isinstance(other, Qobj):
            return self.__rmatmul__(other)
        else:
            res = self.copy()
            res *= other
            return res

    def __imul__(QobjEvo self, other):
        if isinstance(other, (Qobj, QobjEvo)):
            self @= other
        elif isinstance(other, numbers.Number):
            self.elements = [element * other for element in self.elements]
        elif isinstance(other, Coefficient):
            other = _EvoElement(qutip.qeye(self.dims[1]), other)
            self.elements = [element @ other for element in self.elements]
        else:
            return NotImplemented
        return self

    def __truediv__(left, right):
        if isinstance(left, QobjEvo) and isinstance(right, numbers.Number):
            res = left.copy()
            res *= 1 / right
            return res
        return NotImplemented

    def __idiv__(self, other):
        if not isinstance(other, numbers.Number):
            return NotImplemented
        self *= 1 / other
        return self

    def __neg__(self):
        res = self.copy()
        res *= -1
        return res

    ###########################################################################
    # tensor                                                                  #
    ###########################################################################
    def __and__(left, right):
        """
        Syntax shortcut for tensor:
        A & B ==> tensor(A, B)
        """
        return qutip.tensor(left, right)

    ###########################################################################
    # Unary transformation                                                    #
    ###########################################################################
    def trans(self):
        """ Transpose of the quantum object """
        cdef QobjEvo res = self.copy()
        res.elements = [element.linear_map(Qobj.trans)
                        for element in res.elements]
        res._dims = Dimensions(res._dims[0], res._dims[1])
        return res

    def conj(self):
        """Get the element-wise conjugation of the quantum object."""
        cdef QobjEvo res = self.copy()
        res.elements = [element.linear_map(Qobj.conj, True)
                        for element in res.elements]
        return res

    def dag(self):
        """Get the Hermitian adjoint of the quantum object."""
        cdef QobjEvo res = self.copy()
        res.elements = [element.linear_map(Qobj.dag, True)
                        for element in res.elements]
        res._dims = Dimensions(res._dims[0], res._dims[1])
        return res

    def to(self, data_type):
        """
        Convert the underlying data store of all component into the desired
        storage representation.

        The different storage representations available are the "data-layer
        types".  By default, these are :obj:`.Dense`, :obj:`.Dia` and
        :obj:`.CSR`, which respectively construct a dense matrix, diagonal
        sparse matrixand a compressed sparse row one.

        The :obj:`.QobjEvo` is transformed inplace.

        Parameters
        ----------
        data_type : type
            The data-layer type that the data of this :obj:`.Qobj` should be
            converted to.

        Returns
        -------
        None
        """
        return self.linear_map(partial(Qobj.to, data_type=data_type),
                               _skip_check=True)

    def tidyup(self, atol=1e-12):
        """Removes small elements from quantum object."""
        for element in self.elements:
            if type(element) is _ConstantElement:
                element = _ConstantElement(element.qobj(0).tidyup(atol))
            elif type(element) is _EvoElement:
                element = _EvoElement(element.qobj(0).tidyup(atol),
                                      element._coefficient)
        return self

    def linear_map(self, op_mapping, *, _skip_check=False):
        """
        Apply mapping to each Qobj contribution.

        Example:

          ``QobjEvo([sigmax(), coeff]).linear_map(spre)``

        gives the same result has

          ``QobjEvo([spre(sigmax()), coeff])``

        Parameters
        ----------
        op_mapping: callable
            Funtion to apply to each elements.

        Returns
        -------
        :class:`.QobjEvo`
            Modified object

        Notes
        -----
        Does not modify the coefficients, thus ``linear_map(conj)`` would not
        give the the conjugate of the QobjEvo. It's only valid for linear
        transformations.
        """
        if not _skip_check:
            out = op_mapping(self(0))
            if not isinstance(out, Qobj):
                raise TypeError("The op_mapping function must return a Qobj")
        cdef QobjEvo res = self.copy()
        res.elements = [element.linear_map(op_mapping) for element in res.elements]
        res._dims = res.elements[0].qobj(0)._dims
        res.shape = res.elements[0].qobj(0).shape
        res._update_feedback()

        if not _skip_check:
            if res(0) != out:
                raise ValueError("The mapping is not linear")

        return res

    ###########################################################################
    # Cleaning and compress                                                   #
    ###########################################################################
    def _compress_merge_qobj(self, coeff_elements):
        """Merge element with matching qobj:
        ``[A, f1], [A, f2] -> [A, f1+f2]``
        """
        cleaned_elements = []
        # Mimic a dict with Qobj not hashable
        qobjs = []
        coeffs = []
        for element in coeff_elements:
            for i, qobj in enumerate(qobjs):
                if element.qobj(0) == qobj:
                    coeffs[i] = coeffs[i] + element._coefficient
                    break
            else:
                qobjs.append(element.qobj(0))
                coeffs.append(element._coefficient)
        for qobj, coeff in zip(qobjs, coeffs):
            cleaned_elements.append(_EvoElement(qobj, coeff))
        return cleaned_elements

    def compress(self):
        """
        Look for redundance in the :obj:`.QobjEvo` components:

        Constant parts, (:obj:`.Qobj` without :obj:`Coefficient`) will be
        summed.
        Pairs ``[Qobj, Coefficient]`` with the same :obj:`.Qobj` are merged.

        Example:
        ``[[sigmax(), f1], [sigmax(), f2]] -> [[sigmax(), f1+f2]]``

        The :obj:`.QobjEvo` is transformed inplace.

        Returns
        -------
        None
        """
        cte_elements = []
        coeff_elements = []
        func_elements = []
        for element in self.elements:
            if type(element) is _ConstantElement:
                cte_elements.append(element)
            elif type(element) is _EvoElement:
                coeff_elements.append(element)
            else:
                func_elements.append(element)

        cleaned_elements = []
        if len(cte_elements) >= 2:
            # Multiple constant parts
            cleaned_elements.append(_ConstantElement(
                sum(element.qobj(0) for element in cte_elements)))
        else:
            cleaned_elements += cte_elements

        coeff_elements = self._compress_merge_qobj(coeff_elements)
        cleaned_elements += coeff_elements + func_elements

        self.elements = cleaned_elements

    def to_list(QobjEvo self):
        """
        Restore the QobjEvo to a list form.

        Returns
        -------
        list_qevo: list
            The QobjEvo as a list, element are either :obj:`.Qobj` for
            constant parts, ``[Qobj, Coefficient]`` for coefficient based term.
            The original format of the :obj:`Coefficient` is not restored.
            Lastly if the original :obj:`.QobjEvo` is constructed with a
            function returning a Qobj, the term is returned as a pair of the
            original function and args (``dict``).
        """
        out = []
        for element in self.elements:
            if isinstance(element, _ConstantElement):
                out.append(element.qobj(0))
            elif isinstance(element, _EvoElement):
                coeff = element._coefficient
                out.append([element.qobj(0), coeff])
            elif isinstance(element, _FuncElement):
                func = element._func
                args = element._args
                out.append([func, args])
            else:
                out.append([element, {}])
        return out

    ###########################################################################
    # properties                                                              #
    ###########################################################################
    @property
    def num_elements(self):
        """Number of parts composing the system"""
        return len(self.elements)

    @property
    def isconstant(self):
        """Does the system change depending on ``t``"""
        return not any(type(element) is not _ConstantElement
                       for element in self.elements)

    @property
    def isoper(self):
        """Indicates if the system represents an operator."""
        return self._dims.type in ['oper', 'scalar']

    @property
    def issuper(self):
        """Indicates if the system represents a superoperator."""
        return self._dims.type == 'super'

    @property
    def dims(self):
        return self._dims.as_list()

    @property
    def type(self):
        return self._dims.type

    @property
    def superrep(self):
        return self._dims.superrep

    @property
    def isbra(self):
        """Indicates if the system represents a bra state."""
        return self._dims.type in ['bra', 'scalar']

    @property
    def isket(self):
        """Indicates if the system represents a ket state."""
        return self._dims.type in ['ket', 'scalar']

    @property
    def isoperket(self):
        """Indicates if the system represents a operator-ket state."""
        return self._dims.type == 'operator-ket'

    @property
    def isoperbra(self):
        """Indicates if the system represents a operator-bra state."""
        return self._dims.type == 'operator-bra'

    @property
    def dtype(self):
        """
        Type of the data layers of the QobjEvo.
        When different data layers are used, we return the type of the sum of
        the parts.
        """
        part_types = [part.dtype for part in self.elements]
        if (
            part_types[0] is not None
            and all(part == part_types[0] for part in part_types)
        ):
            return part_types[0]
        return self(0).dtype

    ###########################################################################
    # operation methods                                                       #
    ###########################################################################
    def expect(QobjEvo self, object t, state, check_real=True):
        """
        Expectation value of this operator at time ``t`` with the state.

        Parameters
        ----------
        t : float
            Time of the operator to apply.

        state : Qobj
            right matrix of the product

        check_real : bool (True)
            Whether to convert the result to a `real` when the imaginary part
            is smaller than the real part by a dactor of
            ``settings.core['rtol']``.

        Returns
        -------
        expect : float or complex
            ``state.adjoint() @ self @ state`` if ``state`` is a ket.
            ``trace(self @ matrix)`` is ``state`` is an operator or
            operator-ket.
        """
        # TODO: remove reading from `settings` for a typed value when options
        # support property.
        cdef float herm_rtol = settings.core['rtol']
        if not isinstance(state, Qobj):
            raise TypeError("A Qobj state is expected")
        if not (self.isoper or self.issuper):
            raise ValueError("Must be an operator or super operator to compute"
                             " an expectation value")
        if not (
            (self._dims[1] == state._dims[0]) or
            (self.issuper and self._dims[1] == state._dims)
        ):
            raise ValueError("incompatible dimensions " + str(self.dims) +
                             ", " + str(state.dims))
        out = self.expect_data(t, state.data)
        if (
            check_real and
            (out == 0 or (out.real and fabs(out.imag / out.real) < herm_rtol))
        ):
            return out.real
        return out

    cpdef object expect_data(QobjEvo self, object t, Data state):
        """
        Expectation is defined as ``state.adjoint() @ self @ state`` if
        ``state`` is a vector, or ``state`` is an operator and ``self`` is a
        superoperator.  If ``state`` is an operator and ``self`` is an
        operator, then expectation is ``trace(self @ matrix)``.
        """
        if type(state) is Dense:
            return self._expect_dense(t, state)
        cdef _BaseElement part
        cdef object out = 0.
        cdef Data part_data
        cdef object expect_func
        t = self._prepare(t, state)
        if self.issuper:
            if state.shape[1] != 1:
                state = _data.column_stack(state)
            expect_func = _data.expect_super
        else:
            expect_func = _data.expect

        for element in self.elements:
            part = (<_BaseElement> element)
            part_data = part.data(t)
            out += part.coeff(t) * expect_func(part_data, state)
        return out

    cdef double complex _expect_dense(QobjEvo self, double t, Dense state) except *:
        """For Dense state, ``column_stack_dense`` can be done inplace if in
        fortran format."""
        cdef size_t nrow = state.shape[0]
        cdef _BaseElement part
        cdef double complex out = 0., coeff
        cdef Data part_data
        t = self._prepare(t, state)
        if self.issuper:
            if state.shape[1] != 1:
                state = column_stack_dense(state, inplace=state.fortran)
            try:
                for element in self.elements:
                    part = (<_BaseElement> element)
                    coeff = part.coeff(t)
                    part_data = part.data(t)
                    out += coeff * expect_super_data_dense(part_data, state)
            finally:
                if state.fortran:
                    # `state` was reshaped inplace, restore it's original shape
                    column_unstack_dense(state, nrow, inplace=state.fortran)
        else:
            for element in self.elements:
                part = (<_BaseElement> element)
                coeff = part.coeff(t)
                part_data = part.data(t)
                out += coeff * expect_data_dense(part_data, state)
        return out

    def matmul(self, t, state):
        """
        Product of this operator at time ``t`` to the state.
        ``self(t) @ state``

        Parameters
        ----------
        t : float
            Time of the operator to apply.
        state : Qobj
            right matrix of the product

        Returns
        -------
        product : Qobj
            The result product as a Qobj
        """
        if not isinstance(state, Qobj):
            raise TypeError("A Qobj state is expected")

        if self._dims[1] != state._dims[0]:
            raise ValueError("incompatible dimensions " + str(self.dims[1]) +
                             ", " + str(state.dims[0]))

        return Qobj(self.matmul_data(t, state.data),
                    dims=[self._dims[0], state._dims[1]],
                    copy=False
                    )

    cpdef Data matmul_data(QobjEvo self, object t, Data state, Data out=None):
        """Compute ``out += self(t) @ state``"""
        cdef _BaseElement part
        t = self._prepare(t, state)
        if out is None and type(state) is Dense:
            out = dense.zeros(self.shape[0], state.shape[1],
                              (<Dense> state).fortran)
        elif out is None:
            out = _data.zeros[type(state)](self.shape[0], state.shape[1])

        for element in self.elements:
            part = (<_BaseElement> element)
            out = part.matmul_data_t(t, state, out)
        return out


class _Feedback:
    default = None

    def __init__(self):
        raise NotImplementedError("Use subclass")

    def check_consistency(self, dims):
        """
        Raise an error when the dims of the e_ops / state don't match.
        Tell the dims to the feedback for reconstructing the Qobj.
        """



__all__ = ["smesolve", "SMESolver", "ssesolve", "SSESolver"]

from .sode.ssystem import StochasticOpenSystem, StochasticClosedSystem
from .result import MultiTrajResult, Result, ExpectOp
from .multitraj import MultiTrajSolver
from .. import Qobj, QobjEvo
import numpy as np
from functools import partial
from .solver_base import _solver_deprecation
from ._feedback import _QobjFeedback, _DataFeedback, _WeinerFeedback


class StochasticTrajResult(Result):
    def _post_init(self, m_ops=(), dw_factor=(), heterodyne=False):
        super()._post_init()
        self.W = []
        self.m_ops = []
        self.m_expect = []
        self.dW_factor = dw_factor
        self.heterodyne = heterodyne
        for op in m_ops:
            f = self._e_op_func(op)
            self.W.append([0.0])
            self.m_expect.append([])
            self.m_ops.append(ExpectOp(op, f, self.m_expect[-1].append))
            self.add_processor(self.m_ops[-1]._store)

    def add(self, t, state, noise):
        super().add(t, state)
        if noise is not None and self.options["store_measurement"]:
            for i, dW in enumerate(noise):
                self.W[i].append(self.W[i][-1] + dW)

    @property
    def wiener_process(self):
        """
        Wiener processes for each stochastic collapse operators.

        The output shape is
            (len(sc_ops), len(tlist))
        for homodyne detection, and
            (len(sc_ops), 2, len(tlist))
        for heterodyne detection.
        """
        W = np.array(self.W)
        if self.heterodyne:
            W = W.reshape(-1, 2, W.shape[1])
        return W

    @property
    def dW(self):
        """
        Wiener increment for each stochastic collapse operators.

        The output shape is
            (len(sc_ops), len(tlist)-1)
        for homodyne detection, and
            (len(sc_ops), 2, len(tlist)-1)
        for heterodyne detection.
        """
        dw = np.diff(self.W, axis=1)
        if self.heterodyne:
            dw = dw.reshape(-1, 2, dw.shape[1])
        return dw

    @property
    def measurement(self):
        """
        Measurements for each stochastic collapse operators.

        The output shape is
            (len(sc_ops), len(tlist)-1)
        for homodyne detection, and
            (len(sc_ops), 2, len(tlist)-1)
        for heterodyne detection.
        """
        dts = np.diff(self.times)
        m_expect = np.array(self.m_expect)[:, 1:]
        noise = np.einsum(
            "i,ij,j->ij", self.dW_factor, np.diff(self.W, axis=1), (1 / dts)
        )
        if self.heterodyne:
            m_expect = m_expect.reshape(-1, 2, m_expect.shape[1])
            noise = noise.reshape(-1, 2, noise.shape[1])
        return m_expect + noise


class StochasticResult(MultiTrajResult):
    def _post_init(self):
        super()._post_init()

        store_measurement = self.options["store_measurement"]
        keep_runs = self.options["keep_runs_results"]

        if not keep_runs and store_measurement:
            self.add_processor(
                partial(self._reduce_attr, attr="wiener_process")
            )
            self._wiener_process = []
            self.add_processor(partial(self._reduce_attr, attr="dW"))
            self._dW = []

        if not keep_runs and store_measurement:
            self.add_processor(partial(self._reduce_attr, attr="measurement"))
            self._measurement = []

    def _reduce_attr(self, trajectory, attr):
        """
        Add a result attribute to a list when the trajectories are not stored.
        """
        getattr(self, "_" + attr).append(getattr(trajectory, attr))

    def _trajectories_attr(self, attr):
        """
        Get the result associated to the attr, whether the trajectories are
        saved or not.
        """
        if hasattr(self, "_" + attr):
            return getattr(self, "_" + attr)
        elif self.options["keep_runs_results"]:
            return np.array([
                getattr(traj, attr) for traj in self.trajectories
            ])
        return None

    @property
    def measurement(self):
        """
        Measurements for each trajectories and stochastic collapse operators.

        The output shape is
            (ntraj, len(sc_ops), len(tlist)-1)
        for homodyne detection, and
            (ntraj, len(sc_ops), 2, len(tlist)-1)
        for heterodyne detection.
        """
        return self._trajectories_attr("measurement")

    @property
    def dW(self):
        """
        Wiener increment for each trajectories and stochastic collapse
        operators.

        The output shape is
            (ntraj, len(sc_ops), len(tlist)-1)
        for homodyne detection, and
            (ntraj, len(sc_ops), 2, len(tlist)-1)
        for heterodyne detection.
        """
        return self._trajectories_attr("dW")

    @property
    def wiener_process(self):
        """
        Wiener processes for each trajectories and stochastic collapse
        operators.

        The output shape is
            (ntraj, len(sc_ops), len(tlist)-1)
        for homodyne detection, and
            (ntraj, len(sc_ops), 2, len(tlist)-1)
        for heterodyne detection.
        """
        return self._trajectories_attr("wiener_process")


class _StochasticRHS:
    """
    In between object to store the stochastic system.

    It store the Hamiltonian (not Liouvillian when possible), and sc_ops.
    dims and flags are provided to be usable the the base ``Solver`` class.

    We don't want to use the cython rhs (``StochasticOpenSystem``, etc.) since
    the rouchon integrator need the part but does not use the usual drift and
    diffusion computation.
    """

    def __init__(self, issuper, H, sc_ops, c_ops, heterodyne):

        if not isinstance(H, (Qobj, QobjEvo)) or not H.isoper:
            raise TypeError("The Hamiltonian must be am operator")
        self.H = QobjEvo(H)

        if isinstance(sc_ops, (Qobj, QobjEvo)):
            sc_ops = [sc_ops]
        self.sc_ops = [QobjEvo(c_op) for c_op in sc_ops]

        if isinstance(c_ops, (Qobj, QobjEvo)):
            c_ops = [c_ops]
        self.c_ops = [QobjEvo(c_op) for c_op in c_ops]

        if any(not c_op.isoper for c_op in c_ops):
            raise TypeError("c_ops must be operators")

        if any(not c_op.isoper for c_op in sc_ops):
            raise TypeError("sc_ops must be operators")

        self.issuper = issuper
        self.heterodyne = heterodyne
        self._noise_key = None

        if heterodyne:
            sc_ops = []
            for c_op in self.sc_ops:
                sc_ops.append(c_op / np.sqrt(2))
                sc_ops.append(c_op * (-1j / np.sqrt(2)))
            self.sc_ops = sc_ops

        if self.issuper and not self.H.issuper:
            self.dims = [self.H.dims, self.H.dims]
        else:
            self.dims = self.H.dims

    def __call__(self, options):
        if self.issuper:
            return StochasticOpenSystem(
                self.H, self.sc_ops, self.c_ops, options.get("derr_dt", 1e-6)
            )
        else:
            return StochasticClosedSystem(self.H, self.sc_ops)

    def arguments(self, args):
        self.H.arguments(args)
        for c_op in self.c_ops:
            c_op.arguments(args)
        for sc_op in self.sc_ops:
            sc_op.arguments(args)

    def _register_feedback(self, val):
        self.H._register_feedback({"wiener_process": val}, "stochatic solver")
        for c_op in self.c_ops:
            c_op._register_feedback(
                {"WeinerFeedback": val}, "stochatic solver"
            )
        for sc_op in self.sc_ops:
            sc_op._register_feedback(
                {"WeinerFeedback": val}, "stochatic solver"
            )


def smesolve(
    H, rho0, tlist, c_ops=(), sc_ops=(), heterodyne=False, *,
    e_ops=(), args={}, ntraj=500, options=None,
    seeds=None, target_tol=None, timeout=None, **kwargs
):
    """
    Solve stochastic master equation.

    Parameters
    ----------
    H : :obj:`.Qobj`, :obj:`.QobjEvo`, :obj:`.QobjEvo` compatible format.
        System Hamiltonian as a Qobj or QobjEvo for time-dependent
        Hamiltonians. List of [:obj:`.Qobj`, :obj:`.Coefficient`] or callable
        that can be made into :obj:`.QobjEvo` are also accepted.

    rho0 : :class:`.Qobj`
        Initial density matrix or state vector (ket).

    tlist : *list* / *array*
        List of times for :math:`t`.

    c_ops : list of (:obj:`.QobjEvo`, :obj:`.QobjEvo` compatible format), optional
        Deterministic collapse operator which will contribute with a standard
        Lindblad type of dissipation.

    sc_ops : list of (:obj:`.QobjEvo`, :obj:`.QobjEvo` compatible format)
        List of stochastic collapse operators.

    e_ops : : :class:`.qobj`, callable, or list, optional
        Single operator or list of operators for which to evaluate
        expectation values or callable or list of callable.
        Callable signature must be, `f(t: float, state: Qobj)`.
        See :func:`.expect` for more detail of operator expectation.

    args : dict, optional
        Dictionary of parameters for time-dependent Hamiltonians and
        collapse operators.

    ntraj : int, default: 500
        Number of trajectories to compute.

    heterodyne : bool, default: False
        Whether to use heterodyne or homodyne detection.

    seeds : int, SeedSequence, list, optional
        Seed for the random number generator. It can be a single seed used to
        spawn seeds for each trajectory or a list of seeds, one for each
        trajectory. Seeds are saved in the result and they can be reused with::

            seeds=prev_result.seeds

        When using a parallel map, the trajectories can be re-ordered.

    target_tol : {float, tuple, list}, optional
        Target tolerance of the evolution. The evolution will compute
        trajectories until the error on the expectation values is lower than
        this tolerance. The maximum number of trajectories employed is
        given by ``ntraj``. The error is computed using jackknife resampling.
        ``target_tol`` can be an absolute tolerance or a pair of absolute and
        relative tolerance, in that order. Lastly, it can be a list of pairs of
        ``(atol, rtol)`` for each e_ops.

    timeout : float, optional
        Maximum time for the evolution in second. When reached, no more
        trajectories will be computed. Overwrite the option of the same name.

    options : dict, optional
        Dictionary of options for the solver.

        - | store_final_state : bool
          | Whether or not to store the final state of the evolution in the
            result class.
        - | store_states : bool, None
          | Whether or not to store the state vectors or density matrices.
            On `None` the states will be saved if no expectation operators are
            given.
        - | store_measurement: bool
          | Whether to store the measurement and wiener process for each
            trajectories.
        - | keep_runs_results : bool
          | Whether to store results from all trajectories or just store the
            averages.
        - | normalize_output : bool
          | Normalize output state to hide ODE numerical errors.
        - | progress_bar : str {'text', 'enhanced', 'tqdm', ''}
          | How to present the solver progress.
            'tqdm' uses the python module of the same name and raise an error
            if not installed. Empty string or False will disable the bar.
        - | progress_kwargs : dict
          | kwargs to pass to the progress_bar. Qutip's bars use `chunk_size`.
        - | method : str
          | Which stochastic differential equation integration method to use.
            Main ones are {"euler", "rouchon", "platen", "taylor1.5_imp"}
        - | map : str {"serial", "parallel", "loky", "mpi"}
          | How to run the trajectories. "parallel" uses the multiprocessing
            module to run in parallel while "loky" and "mpi" use the "loky" and
            "mpi4py" modules to do so.
        - | num_cpus : NoneType, int
          | Number of cpus to use when running in parallel. ``None`` detect the
            number of available cpus.
        - | dt : float
          | The finite steps lenght for the Stochastic integration method.
            Default change depending on the integrator.

        Additional options are listed under
        `options <./classes.html#qutip.solver.stochastic.SMESolver.options>`__.
        More options may be available depending on the selected
        differential equation integration method, see
        `SIntegrator <./classes.html#classes-sode>`_.

    Returns
    -------

    output: :class:`.Result`
        An instance of the class :class:`.Result`.
    """
    options = _solver_deprecation(kwargs, options, "stoc")
    H = QobjEvo(H, args=args, tlist=tlist)
    c_ops = [QobjEvo(c_op, args=args, tlist=tlist) for c_op in c_ops]
    sc_ops = [QobjEvo(c_op, args=args, tlist=tlist) for c_op in sc_ops]
    sol = SMESolver(
        H, sc_ops, c_ops=c_ops, options=options, heterodyne=heterodyne
    )
    return sol.run(
        rho0, tlist, ntraj, e_ops=e_ops,
        seeds=seeds, target_tol=target_tol, timeout=timeout,
    )


def ssesolve(
    H, psi0, tlist, sc_ops=(), heterodyne=False, *,
    e_ops=(), args={}, ntraj=500, options=None,
    seeds=None, target_tol=None, timeout=None, **kwargs
):
    """
    Solve stochastic Schrodinger equation.

    Parameters
    ----------
    H : :obj:`.Qobj`, :obj:`.QobjEvo`, :obj:`.QobjEvo` compatible format.
        System Hamiltonian as a Qobj or QobjEvo for time-dependent
        Hamiltonians. List of [:obj:`.Qobj`, :obj:`.Coefficient`] or callable
        that can be made into :obj:`.QobjEvo` are also accepted.

    psi0 : :class:`.Qobj`
        Initial state vector (ket).

    tlist : *list* / *array*
        List of times for :math:`t`.

    sc_ops : list of (:obj:`.QobjEvo`, :obj:`.QobjEvo` compatible format)
        List of stochastic collapse operators.

    e_ops : :class:`.qobj`, callable, or list, optional
        Single operator or list of operators for which to evaluate
        expectation values or callable or list of callable.
        Callable signature must be, `f(t: float, state: Qobj)`.
        See :func:`expect` for more detail of operator expectation.

    args : dict, optional
        Dictionary of parameters for time-dependent Hamiltonians and
        collapse operators.

    ntraj : int, default: 500
        Number of trajectories to compute.

    heterodyne : bool, default: False
        Whether to use heterodyne or homodyne detection.

    seeds : int, SeedSequence, list, optional
        Seed for the random number generator. It can be a single seed used to
        spawn seeds for each trajectory or a list of seeds, one for each
        trajectory. Seeds are saved in the result and they can be reused with::

            seeds=prev_result.seeds

    target_tol : {float, tuple, list}, optional
        Target tolerance of the evolution. The evolution will compute
        trajectories until the error on the expectation values is lower than
        this tolerance. The maximum number of trajectories employed is
        given by ``ntraj``. The error is computed using jackknife resampling.
        ``target_tol`` can be an absolute tolerance or a pair of absolute and
        relative tolerance, in that order. Lastly, it can be a list of pairs of
        (atol, rtol) for each e_ops.

    timeout : float, optional
        Maximum time for the evolution in second. When reached, no more
        trajectories will be computed. Overwrite the option of the same name.

    options : dict, optional
        Dictionary of options for the solver.

        - | store_final_state : bool
          | Whether or not to store the final state of the evolution in the
            result class.
        - | store_states : bool, None
          | Whether or not to store the state vectors or density matrices.
            On `None` the states will be saved if no expectation operators are
            given.
        - | store_measurement: bool
            Whether to store the measurement and wiener process, or brownian
            noise for each trajectories.
        - | keep_runs_results : bool
          | Whether to store results from all trajectories or just store the
            averages.
        - | normalize_output : bool
          | Normalize output state to hide ODE numerical errors.
        - | progress_bar : str {'text', 'enhanced', 'tqdm', ''}
          | How to present the solver progress.
            'tqdm' uses the python module of the same name and raise an error
            if not installed. Empty string or False will disable the bar.
        - | progress_kwargs : dict
          | kwargs to pass to the progress_bar. Qutip's bars use `chunk_size`.
        - | method : str
          | Which stochastic differential equation integration method to use.
            Main ones are {"euler", "rouchon", "platen", "taylor1.5_imp"}
        - | map : str {"serial", "parallel", "loky", "mpi"}
          | How to run the trajectories. "parallel" uses the multiprocessing
            module to run in parallel while "loky" and "mpi" use the "loky" and
            "mpi4py" modules to do so.
        - | num_cpus : NoneType, int
          | Number of cpus to use when running in parallel. ``None`` detect the
            number of available cpus.
        - | dt : float
          | The finite steps lenght for the Stochastic integration method.
            Default change depending on the integrator.

        Additional options are listed under
        `options <./classes.html#qutip.solver.stochastic.SSESolver.options>`__.
        More options may be available depending on the selected
        differential equation integration method, see
        `SIntegrator <./classes.html#classes-sode>`_.

    Returns
    -------

    output: :class:`.Result`
        An instance of the class :class:`.Result`.
    """
    options = _solver_deprecation(kwargs, options, "stoc")
    H = QobjEvo(H, args=args, tlist=tlist)
    sc_ops = [QobjEvo(c_op, args=args, tlist=tlist) for c_op in sc_ops]
    sol = SSESolver(H, sc_ops, options=options, heterodyne=heterodyne)
    return sol.run(
        psi0, tlist, ntraj, e_ops=e_ops,
        seeds=seeds, target_tol=target_tol, timeout=timeout,
    )


class StochasticSolver(MultiTrajSolver):
    """
    Generic stochastic solver.
    """

    name = "StochasticSolver"
    _resultclass = StochasticResult
    _avail_integrators = {}
    system = None
    _open = None
    solver_options = {
        "progress_bar": "text",
        "progress_kwargs": {"chunk_size": 10},
        "store_final_state": False,
        "store_states": None,
        "keep_runs_results": False,
        "normalize_output": False,
        "map": "serial",
        "mpi_options": {},
        "num_cpus": None,
        "bitgenerator": None,
        "method": "platen",
        "store_measurement": False,
    }

    def __init__(self, H, sc_ops, heterodyne, *, c_ops=(), options=None):
        self.options = options
        self._heterodyne = heterodyne
        if self.name == "ssesolve" and c_ops:
            raise ValueError("c_ops are not supported by ssesolve.")

        rhs = _StochasticRHS(self._open, H, sc_ops, c_ops, heterodyne)
        self.rhs = rhs
        self.system = rhs
        self.options = options
        self.seed_sequence = np.random.SeedSequence()
        self._integrator = self._get_integrator()
        self._state_metadata = {}
        self.stats = self._initialize_stats()

        if heterodyne:
            self._m_ops = []
            for op in sc_ops:
                self._m_ops += [op + op.dag(), -1j * (op - op.dag())]
            self._dW_factors = np.ones(len(sc_ops) * 2) * 2**0.5
        else:
            self._m_ops = [op + op.dag() for op in sc_ops]
            self._dW_factors = np.ones(len(sc_ops))

    @property
    def heterodyne(self):
        return self._heterodyne

    @property
    def m_ops(self):
        return self._m_ops

    @m_ops.setter
    def m_ops(self, new_m_ops):
        """
        Measurements operators.

        Default are:

            m_ops = sc_ops + sc_ops.dag()

        for homodyne detection, and

            m_ops = sc_ops + sc_ops.dag(), -1j*(sc_ops - sc_ops.dag())

        for heterodyne detection.

        Measurements opput is computed as:

            expect(m_ops_i, state(t)) + dW_i / dt * dW_factors

        Where ``dW`` follows a gaussian distribution with norm 0 and derivation
        of ``dt**0.5``. ``dt`` is the time difference between step in the
        ``tlist``.

        ``m_ops`` can be overwritten, but the number of operators must be
        constant.
        """
        if len(new_m_ops) != len(self.m_ops):
            if self.heterodyne:
                raise ValueError(
                    f"2 `m_ops` per `sc_ops`, {len(self.rhs.sc_ops)} operators"
                    " are expected for heterodyne measurement."
                )
            else:
                raise ValueError(
                    f"{len(self.rhs.sc_ops)} measurements "
                    "operators are expected."
                )
        if not all(
            isinstance(op, Qobj) and op.dims == self.rhs.sc_ops[0].dims
            for op in new_m_ops
        ):
            raise ValueError(
                "m_ops must be Qobj with the same dimensions"
                " as the Hamiltonian"
            )
        self._m_ops = new_m_ops

    @property
    def dW_factors(self):
        return self._dW_factors

    @dW_factors.setter
    def dW_factors(self, new_dW_factors):
        """
        Scaling of the noise on the measurements.
        Default are ``1`` for homodyne and ``sqrt(1/2)`` for heterodyne.
        ``dW_factors`` must be a list of the same length as ``m_ops``.
        """
        if len(new_dW_factors) != len(self._dW_factors):
            if self.heterodyne:
                raise ValueError(
                    f"2 `dW_factors` per `sc_ops`, {len(self.rhs.sc_ops)} "
                    "values are expected for heterodyne measurement."
                )
            else:
                raise ValueError(
                    f"{len(self.rhs.sc_ops)} dW_factors are expected."
                )
        self._dW_factors = new_dW_factors

    def _run_one_traj(self, seed, state, tlist, e_ops):
        """
        Run one trajectory and return the result.
        """
        result = StochasticTrajResult(
            e_ops,
            self.options,
            m_ops=self.m_ops,
            dw_factor=self.dW_factors,
            heterodyne=self.heterodyne,
        )
        generator = self._get_generator(seed)
        self._integrator.set_state(tlist[0], state, generator)
        state_t = self._restore_state(state, copy=False)
        result.add(tlist[0], state_t, None)
        for t in tlist[1:]:
            t, state, noise = self._integrator.integrate(t, copy=False)
            state_t = self._restore_state(state, copy=False)
            result.add(t, state_t, noise)
        return seed, result

    @classmethod
    def avail_integrators(cls):
        if cls is StochasticSolver:
            return cls._avail_integrators.copy()
        return {
            **StochasticSolver.avail_integrators(),
            **cls._avail_integrators,
        }

    @property
    def options(self):
        """
        Options for stochastic solver:

        store_final_state: bool, default: False
            Whether or not to store the final state of the evolution in the
            result class.

        store_states: None, bool, default: None
            Whether or not to store the state vectors or density matrices.
            On `None` the states will be saved if no expectation operators are
            given.

        store_measurement: bool, default: False
            Whether to store the measurement for each trajectories.
            Storing measurements will also store the wiener process, or
            brownian noise for each trajectories.

        progress_bar: str {'text', 'enhanced', 'tqdm', ''}, default: "text"
            How to present the solver progress. 'tqdm' uses the python module
            of the same name and raise an error if not installed. Empty string
            or False will disable the bar.

        progress_kwargs: dict, default: {"chunk_size":10}
            Arguments to pass to the progress_bar. Qutip's bars use
            ``chunk_size``.

        keep_runs_results: bool, default: False
          Whether to store results from all trajectories or just store the
          averages.

        normalize_output: bool
            Normalize output state to hide ODE numerical errors.

        method: str, default: "platen"
            Which differential equation integration method to use.

        map: str {"serial", "parallel", "loky", "mpi"}, default: "serial"
            How to run the trajectories. "parallel" uses the multiprocessing
            module to run in parallel while "loky" and "mpi" use the "loky" and
            "mpi4py" modules to do so.

        mpi_options: dict, default: {}
            Only applies if map is "mpi". This dictionary will be passed as
            keyword arguments to the `mpi4py.futures.MPIPoolExecutor`
            constructor. Note that the `max_workers` argument is provided
            separately through the `num_cpus` option.

        num_cpus: None, int, default: None
            Number of cpus to use when running in parallel. ``None`` detect the
            number of available cpus.

        bitgenerator: {None, "MT19937", "PCG64DXSM", ...}, default: None
            Which of numpy.random's bitgenerator to use. With ``None``, your
            numpy version's default is used.
        """
        return self._options

    @options.setter
    def options(self, new_options):
        MultiTrajSolver.options.fset(self, new_options)

    @classmethod
    def WeinerFeedback(cls, default=None):
        """
        Weiner function of the trajectory argument for time dependent systems.

        When used as an args:

            ``QobjEvo([op, func], args={"W": SMESolver.WeinerFeedback()})``

        The ``func`` will receive a function as ``W`` that return an array of
        wiener processes values at ``t``. The wiener process for the i-th
        sc_ops is the i-th element for homodyne detection and the (2i, 2i+1)
        pairs of process in heterodyne detection. The process is a step
        function with step of length ``options["dt"]``.

        .. note::

            WeinerFeedback can't be added to a running solver when updating
            arguments between steps: ``solver.step(..., args={})``.

        Parameters
        ----------
        default : callable, optional
            Default function used outside the solver.
            When not passed, a function returning ``np.array([0])`` is used.

        """
        return _WeinerFeedback(default)

    @classmethod
    def StateFeedback(cls, default=None, raw_data=False):
        """
        State of the evolution to be used in a time-dependent operator.

        When used as an args:

            ``QobjEvo([op, func], args={"state": SMESolver.StateFeedback()})``

        The ``func`` will receive the density matrix as ``state`` during the
        evolution.

        .. note::

            Not supported by the ``rouchon`` mehtod.

        Parameters
        ----------
        default : Qobj or qutip.core.data.Data, default : None
            Initial value to be used at setup of the system.

        raw_data : bool, default : False
            If True, the raw matrix will be passed instead of a Qobj.
            For density matrices, the matrices can be column stacked or square
            depending on the integration method.

        """
        if raw_data:
            return _DataFeedback(default, open=cls._open)
        return _QobjFeedback(default, open=cls._open)


class SMESolver(StochasticSolver):
    r"""
    Stochastic Master Equation Solver.

    Parameters
    ----------
    H : :obj:`.Qobj`, :obj:`.QobjEvo`, :obj:`.QobjEvo` compatible format.
        System Hamiltonian as a Qobj or QobjEvo for time-dependent
        Hamiltonians. List of [:obj:`.Qobj`, :obj:`.Coefficient`] or callable
        that can be made into :obj:`.QobjEvo` are also accepted.

    sc_ops : list of (:obj:`.QobjEvo`, :obj:`.QobjEvo` compatible format)
        List of stochastic collapse operators.

    heterodyne : bool, default: False
        Whether to use heterodyne or homodyne detection.

    options : dict, optional
        Options for the solver, see :obj:`SMESolver.options` and
        `SIntegrator <./classes.html#classes-sode>`_ for a list of all options.
    """
    name = "smesolve"
    _avail_integrators = {}
    _open = True
    solver_options = {
        "progress_bar": "text",
        "progress_kwargs": {"chunk_size": 10},
        "store_final_state": False,
        "store_states": None,
        "keep_runs_results": False,
        "normalize_output": False,
        "map": "serial",
        "mpi_options": {},
        "num_cpus": None,
        "bitgenerator": None,
        "method": "platen",
        "store_measurement": False,
    }


class SSESolver(StochasticSolver):
    r"""
    Stochastic Schrodinger Equation Solver.

    Parameters
    ----------
    H : :obj:`.Qobj`, :obj:`.QobjEvo`, :obj:`.QobjEvo` compatible format.
        System Hamiltonian as a Qobj or QobjEvo for time-dependent
        Hamiltonians. List of [:obj:`.Qobj`, :obj:`.Coefficient`] or callable
        that can be made into :obj:`.QobjEvo` are also accepted.

    c_ops : list of (:obj:`.QobjEvo`, :obj:`.QobjEvo` compatible format)
        Deterministic collapse operator which will contribute with a standard
        Lindblad type of dissipation.

    sc_ops : list of (:obj:`.QobjEvo`, :obj:`.QobjEvo` compatible format)
        List of stochastic collapse operators.

    heterodyne : bool, default: False
        Whether to use heterodyne or homodyne detection.

    options : dict, optional
        Options for the solver, see :obj:`SSESolver.options` and
        `SIntegrator <./classes.html#classes-sode>`_ for a list of all options.
    """
    name = "ssesolve"
    _avail_integrators = {}
    _open = False
    solver_options = {
        "progress_bar": "text",
        "progress_kwargs": {"chunk_size": 10},
        "store_final_state": False,
        "store_states": None,
        "keep_runs_results": False,
        "normalize_output": False,
        "map": "serial",
        "mpi_options": {},
        "num_cpus": None,
        "bitgenerator": None,
        "method": "platen",
        "store_measurement": False,
    }



import pytest
pytest.importorskip("IPython")

from qutip.ipynbtools import version_table

@pytest.mark.parametrize('verbose', [False, True])
def test_version_table(verbose):
    html_data = version_table(verbose=verbose).data
    assert "<th>Software</th>" in html_data
    assert "<th>Version</th>" in html_data
    assert "<td>QuTiP</td>" in html_data
    assert "<td>Numpy</td>" in html_data
    assert "<td>SciPy</td>" in html_data
    assert "<td>matplotlib</td>" in html_data
    assert "<td>IPython</td>" in html_data
    if verbose:
        assert "<td>Installation path</td>" in html_data
        if pytest.importorskip("getpass") is not None:
            assert "<td>User</td>" in html_data


@pytest.mark.skipif(not pytest.importorskip("Cython"), reason="cython not installed")
def test_version_table_with_cython():
    html_data = version_table().data
    assert "<td>Cython</td>" in html_data



#cython: language_level=3

cdef extern from "src/zspmv_openmp.hpp" nogil:
    void zspmvpy_openmp(double complex *data, int *ind, int *ptr, double complex *vec,
                double complex a, double complex *out, int nrows, int nthr)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void spmvpy_openmp(complex * data, int * ind, int * ptr,
                               complex * vec, complex a, complex * out,
                               unsigned int nrows, unsigned int nthr):
    zspmvpy_openmp(data, ind, ptr, vec, a, out, nrows, nthr)
